#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "LiveStorageAdapter.h"

#include <arviz_godot/lib/avi/core/Constants.h>

avi::connection::LiveStorageAdapter::LiveStorageAdapter(avi::connection::Component& component) :
    component(component)
{
}

avi::connection::StorageAdapter::LayerUpdates
avi::connection::LiveStorageAdapter::getLayerUpdates()
{
    try
    {
        if (!component.isConnected())
            return {};

        armarx::viz::data::LayerUpdates updates =
            component.getStorageInterface()->pullUpdatesSinceAndSendInteractions(lastRevision,
                                                                                 interactions);
        lastRevision = updates.revision;
        interactions = {};
        return {updates,false};
    }
    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while pulling updates from storage: " << e.what();
        return {};
    }
}

std::vector<std::string>
avi::connection::LiveStorageAdapter::getAllReplays()
{
    try
    {
        armarx::viz::data::RecordingsInfo recordingInfo =
            component.getStorageInterface()->getAllRecordings();

        std::vector<armarx::viz::data::Recording> replays = recordingInfo.recordings;
        recordingPath = recordingInfo.recordingsPath;
        std::vector<std::string> replayNames;
        replayNames.reserve(replays.size());

        for (const auto& replay : replays)
            replayNames.push_back(replay.id);

        return replayNames;
    }
    catch (IceUtil::Exception& e)
    {
        return {};
    }
}

void
avi::connection::LiveStorageAdapter::setInteraction(
    std::vector<avi::connection::Interaction> newInteractions)
{
    DETOOLIFY;
    for (auto interaction : newInteractions)
    {
        armarx::viz::data::InteractionFeedback feedback;
        switch (interaction.getType())
        {
            case InteractionType::OPTION:
                feedback.type = armarx::viz::data::InteractionFeedbackType::CONTEXT_MENU_CHOSEN;
                feedback.chosenContextMenuEntry = interaction.getOption().value();
                break;
            case InteractionType::SELECTION:
                feedback.type = armarx::viz::data::InteractionFeedbackType::SELECT;
                break;
            case InteractionType::DESELECTION:
                feedback.type = armarx::viz::data::InteractionFeedbackType::DESELECT;
                break;
            case InteractionType::MANIPULATION:
                feedback.type = armarx::viz::data::InteractionFeedbackType::TRANSFORM;
                switch (interaction.getTransformation())
                {
                    case avi::connection::Interaction::TRANSFORMATION_BEGIN:
                        feedback.type |=
                            armarx::viz::data::InteractionFeedbackType::TRANSFORM_BEGIN_FLAG;
                        break;

                    case avi::connection::Interaction::TRANSFORMATION_DURING:
                        feedback.type |=
                            armarx::viz::data::InteractionFeedbackType::TRANSFORM_DURING_FLAG;
                        break;

                    case avi::connection::Interaction::TRANSFORMATION_END:
                        feedback.type |=
                            armarx::viz::data::InteractionFeedbackType::TRANSFORM_END_FLAG;
                        break;
                }
                break;
        }
        avi::core::Element* element = interaction.getElement();
        feedback.scale = convertVector(interaction.getScale());
        feedback.element = element->getName();
        feedback.layer = element->getLayer()->getName();
        feedback.revision = lastRevision;
        feedback.component = element->getLayer()->getComponent();

        armarx::viz::data::GlobalPose pose;

        // converting m to mm
        armarx::Vector3f translation = convertVector(interaction.getTranslation() *
                                                     avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

        pose.x = translation.e0;
        pose.y = translation.e1;
        pose.z = translation.e2;

        pose.qx = interaction.getRotation().x;
        pose.qy = interaction.getRotation().y;
        pose.qz = interaction.getRotation().z;
        pose.qw = interaction.getRotation().w;

        feedback.transformation = pose;

        interactions.push_back(feedback);
    }
}

void
avi::connection::LiveStorageAdapter::setup()
{
    DETOOLIFY;
}

armarx::Vector3f
avi::connection::LiveStorageAdapter::convertVector(godot::Vector3 vec)
{
    armarx::Vector3f out = {.e0 = vec.x, .e1 = vec.y, .e2 = vec.z};
    return out;
}

bool avi::connection::LiveStorageAdapter::finishedProcessing() {
    return true;
}