#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ReplayStorageAdapter.h"

#include <utility>
#include <godot_cpp/classes/os.hpp>

avi::connection::ReplayStorageAdapter::ReplayStorageAdapter(
    avi::connection::Component& component,
    std::shared_ptr<avi::connection::Clock> clock,
    std::string replay) :
    component(component), clock(std::move(clock)), activeReplay(std::move(replay))
{
    DETOOLIFY;
}

avi::connection::StorageAdapter::LayerUpdates
avi::connection::ReplayStorageAdapter::getLayerUpdates() {
    try {
        auto time = std::chrono::duration_cast<std::chrono::microseconds>(
                std::chrono::steady_clock::now() - last)
                .count();


        armarx::viz::data::LayerUpdates updates;

        if(clock->replayClearScene()) {
            return {updates, true};
        }

        if(clock->hasTargetTimeUpdate()) {
            batchLoadQueue.clear();
            if(asyncBatchLoading) {
                discardLoadingBatch = true;
            }
        }

        //triggered by user setting target revision
        if(clock->hasRevisionUpdate()) {
            skipToRevision(clock->getTargetRevision());
            batchLoadQueue.clear();
            if(asyncBatchLoading) {
                discardLoadingBatch = true;
            }
        }

        int64_t wantedTimeStamp = clock->getTargetTime() + clock->getTimeFrame().begin;
        int64_t index = getBatchIndex(wantedTimeStamp);

        if(not buffering) {
            clock->advance(time, false);
            clock->setLoadingBatches(0);
        }

        if (index == -1) {
            return {updates, false};
        }

        updateBatchLoadQueue(index);
        updateBatchCache(index);
        asyncLoadBatch();
        last = std::chrono::steady_clock::now();

        for (auto invalidBatch: invalidBatches) {
            if (invalidBatch == index) {
                return {updates, false};
            }
        }

        return getUpdates(index, wantedTimeStamp);
    }
    catch (IceUtil::Exception &e) {
        ARMARX_ERROR << "Error while pulling updates from storage: " << e.what();
        return {};
    }
}

avi::connection::StorageAdapter::LayerUpdates avi::connection::ReplayStorageAdapter::getUpdates(int index, int64_t wantedTimeStamp) {

    armarx::viz::data::LayerUpdates updates;

    if (batchCache.find(index) == batchCache.end()) {
        last = std::chrono::steady_clock::now();
        buffering = true;
        clock->setLoadingBatches(batchLoadQueue.size());
        return {updates, false};
    }

    buffering = false;


    armarx::viz::data::RecordingBatch currentBatch = batchCache.at(index).data;
    batchCache.at(index).lastUsed = IceUtil::Time::now();

    for(auto update: currentBatch.initialState) {
        updates.updates.push_back(update.update);
    }

    size_t i =
            binarySearch(currentBatch.updates, 0, currentBatch.updates.size(), wantedTimeStamp);
    clock->updateTargetRevision(currentBatch.updates[i].revision);
    for (size_t j = 0; j < i; j++) {
        updates.updates.push_back(currentBatch.updates[j].update);
    }

    return {updates, false};
}

size_t
avi::connection::ReplayStorageAdapter::binarySearch(
        const std::vector<armarx::viz::data::TimestampedLayerUpdate> &updates,
        size_t l,
        size_t r,
        int64_t timeStamp) {
    if (r >= l) {
        size_t mid = l + (r - l) / 2;
        if ((updates[mid].timestampInMicroseconds <= timeStamp) &&
            (mid + 1 == updates.size() || updates[mid + 1].timestampInMicroseconds > timeStamp)) {
            return mid;
        } else if (updates[mid].timestampInMicroseconds > timeStamp) {
            return binarySearch(updates, l, mid - 1, timeStamp);
        } else {
            return binarySearch(updates, mid + 1, r, timeStamp);
        }
    } else {
        return -1;
    }
}

std::vector<std::string>
avi::connection::ReplayStorageAdapter::getAllReplays() {
    try {
        std::vector<armarx::viz::data::Recording> replays =
                component.getStorageInterface()->getAllRecordings().recordings;
        std::vector<std::string> replayNames;
        replayNames.reserve(replays.size());

        for (const auto &replay: replays) {
            replayNames.push_back(replay.id);
        }

        return replayNames;
    }
    catch (IceUtil::Exception &e) {
        ARMARX_ERROR << "Error while pulling replay names: " << e.what();
        return {};
    }
}

void
avi::connection::ReplayStorageAdapter::setup()
{
    DETOOLIFY;
    try {
        auto recordings = component.getStorageInterface()->getAllRecordings().recordings;

        if (recordings.empty())
            return;

        size_t index = 0;
        while (recordings[index].id != activeReplay) {
            index++;
        }
        batchCount = recordings[index].batchHeaders.size();
        clock->setTimeFrame(TimeFrame(recordings[index].firstTimestampInMicroSeconds,
                                      recordings[index].lastTimestampInMicroSeconds));
        clock->setRevisionFrame(recordings[index].firstRevision, recordings[index].lastRevision);

        for (size_t i = 0; i < recordings[index].batchHeaders.size(); i++) {
            batches.push_back(batchInfo{recordings[index].batchHeaders[i].firstTimestampInMicroSeconds,
                                        recordings[index].batchHeaders[i].lastTimestampInMicroSeconds,
                                        recordings[index].batchHeaders[i].firstRevision,
                                        recordings[index].batchHeaders[i].lastRevision,
                                        static_cast<int64_t>(i)});
        }

        clock->setLoadingBatches(batchLoadQueue.size());
        TimestampedRecordingBatch timestampedRecordingBatch;
        timestampedRecordingBatch.data =
                component.getStorageInterface()->getRecordingBatch(activeReplay, 0);
        timestampedRecordingBatch.lastUsed = IceUtil::Time::now();

        batchCache.insert(std::make_pair<size_t, TimestampedRecordingBatch>(
                0, std::move(timestampedRecordingBatch)));
        clock->setAvailableTimeFrame(
                TimeFrame(recordings[index].batchHeaders[0].firstTimestampInMicroSeconds,
                          recordings[index].batchHeaders[0].lastTimestampInMicroSeconds));
        clock->setLoadingBatches(0);

        callbackData = new ArVizWidgetBatchCallback();
        callbackData->this_ = this;
        callback = armarx::viz::newCallback_StorageInterface_getRecordingBatch(
                callbackData,
                &ArVizWidgetBatchCallback::onSuccess,
                &ArVizWidgetBatchCallback::onFailure);
    }

    catch (IceUtil::Exception &e) {
        ARMARX_ERROR << "Error while loading replay information: " << e.what();
    }
}

void
avi::connection::ReplayStorageAdapter::setInteraction(
        std::vector<avi::connection::Interaction> newInteractions) {
    DETOOLIFY;
}

void
avi::connection::ReplayStorageAdapter::asyncLoadBatch() {
    DETOOLIFY;
    if (batchLoadQueue.empty() || asyncBatchLoading) return;

    asyncBatchLoading = true;
    int64_t nextBatchIndex = batchLoadQueue.front();

    callbackResult = component.getStorageInterface()->begin_getRecordingBatch(
            activeReplay, nextBatchIndex, callback);
}

void
avi::connection::ReplayStorageAdapter::updateBatchLoadQueue(int index) {
    DETOOLIFY;

    int64_t batchesSize = batches.size();
    std::vector<int64_t> requestedBatches;
    int sign = 1;

    if (clock->getSpeed() < 0) {
        sign = -1;
    }

    for (int64_t i = -(sign * BATCHES_BACK); i != (sign * (BATCHES_FRONT)); i = i + sign) {
        int64_t batchIndex = (batchesSize + ((index + i) % batchesSize)) % batchesSize;
        requestedBatches.push_back(batchIndex);
    }

    //only add necessary batches to load queue
    for (auto batchIndex: requestedBatches) {
        bool alreadyLoaded = false;
        for (int64_t entry: batchLoadQueue) {
            if (batchIndex == entry) {
                alreadyLoaded = true;
            }
        }

        for (auto &entry: batchCache) {
            if (batchIndex == entry.first) {
                alreadyLoaded = true;
            }
        }

        for (auto entry: invalidBatches) {
            if (batchIndex == entry) {
                alreadyLoaded = true;
            }
        }

        if (!alreadyLoaded) {
            batchLoadQueue.push_back(batchIndex);
        }
    }
}


void
avi::connection::ReplayStorageAdapter::onGetBatchAsync(
        const armarx::viz::data::RecordingBatch &batch) {
    batchMutex.lock();

    if (!discardLoadingBatch) {
        auto batchIndex = batchLoadQueue.begin();
        batchLoadQueue.erase(batchIndex);

        asyncReceivedBatches.emplace_back(AsyncBatch{batch, *batchIndex});
    }

    discardLoadingBatch = false;
    asyncBatchLoading = false;

    batchMutex.unlock();
}

void avi::connection::ReplayStorageAdapter::updateBatchCache(int targetIndex) {

    if (asyncReceivedBatches.empty()) return;


    std::vector<armarx::viz::data::RecordingBatch> receivedBatches;

    batchMutex.lock();
    auto iterator = asyncReceivedBatches.begin();
    while (iterator != asyncReceivedBatches.end()) {
        int batchHeaderIndex = iterator.base()->batch.header.index;
        if (!(batchHeaderIndex < 0 || batchHeaderIndex >= batchCount)) {
            receivedBatches.emplace_back(iterator.base()->batch);
        } else {
            invalidBatches.emplace_back(iterator->batchIndex);

            if(invalidBatches.size() < 6) {
                godot::PackedStringArray poolStringArray;
                poolStringArray.push_back(
                        "Invalid batch found: " + godot::String(std::to_string(iterator->batchIndex).c_str()));
                poolStringArray.push_back("Check ArVizStorage warning for more information.");
                godot::OS::get_singleton()->execute("notify-send", poolStringArray);
            } else if (invalidBatches.size() == 6) {
                godot::PackedStringArray poolStringArray;
                poolStringArray.push_back(
                        "Recording contains more than 5 invalid batches.");
                poolStringArray.push_back("Check ArVizStorage warning for more information.");
                godot::OS::get_singleton()->execute("notify-send", poolStringArray);
            }
        }
        iterator = asyncReceivedBatches.erase(iterator);
    }
    batchMutex.unlock();

    for (auto batch: receivedBatches) {
        TimestampedRecordingBatch batchEntry;

        batchEntry.data = batch;
        batchEntry.lastUsed = IceUtil::Time::now();

        batchCache.insert(std::make_pair<size_t, TimestampedRecordingBatch>(
                batchEntry.data.header.index, std::move(batchEntry)));
    }

    //remove old batches
    if (batchCache.size() <= BATCH_CACHE_SIZE) {
        return;
    }

    // Remove the entry with the oldest last used timestamp
    auto oldestIter = batchCache.begin();
    for (auto iter = batchCache.begin(); iter != batchCache.end(); ++iter) {
        TimestampedRecordingBatch &entry = iter->second;
        if (entry.lastUsed < oldestIter->second.lastUsed) {
            oldestIter = iter;
        }
    }

    if((lastSentBatch < targetIndex
    && oldestIter->first >= lastSentBatch
    && oldestIter->first <= targetIndex)) {
        return;
    }

    if(lastSentBatch > targetIndex
    && oldestIter->first <= lastSentBatch
    && oldestIter->first >= targetIndex) {
        return;
    }

    batchCache.erase(oldestIter);
}

void avi::connection::ReplayStorageAdapter::skipToRevision(long revision) {
    long timestamp = -1;
    int batchIndex = -1;
    for (auto batch: batches) {
        if (batch.firstRevision <= revision && batch.lastRevision > revision) {
            batchIndex = batch.batchIndex;
            break;
        }
    }

    TimestampedRecordingBatch recordingBatch;
    bool found = false;
    for (auto batch: batchCache) {
        if (batch.second.data.header.index == batchIndex) {
            recordingBatch = batch.second;
            found = true;
            break;
        }
    }

    if (not found) {
        clock->setLoadingBatches(batchLoadQueue.size());
        recordingBatch.data = component.getStorageInterface()->getRecordingBatch(activeReplay, batchIndex);
        recordingBatch.lastUsed = IceUtil::Time::now();
        batchCache.insert(std::make_pair<size_t, TimestampedRecordingBatch>(
                batchIndex, std::move(recordingBatch)));
        clock->setLoadingBatches(0);

        for (auto batch: batchCache) {
            if (batch.second.data.header.index == batchIndex) {
                recordingBatch = batch.second;
                break;
            }
        }
    }

    for (auto update: recordingBatch.data.updates) {
        if (update.revision == revision) {
            timestamp = update.timestampInMicroseconds;
        }
    }

    clock->advance(timestamp - clock->getTargetTime() - clock->getTimeFrame().begin, true);
}

bool avi::connection::ReplayStorageAdapter::finishedProcessing() {
    return !asyncBatchLoading;
}

int64_t avi::connection::ReplayStorageAdapter::getBatchIndex(int64_t timeStamp) {
    int64_t index = -1;

    if (timeStamp > clock->getTimeFrame().end ||
        timeStamp < clock->getTimeFrame().begin) {
        return index;
    }

    // get index of batch containing wantedTimeStamp
    for (auto &batch: batches) {
        if (batch.firstTimestamp <= timeStamp && batch.lastTimestamp >= timeStamp){
            index = batch.batchIndex;
            break;
        }
    }

    return index;
}
