#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "UserInterface.h"

#include <godot_cpp/classes/input.hpp>
#include <godot_cpp/classes/input_event_key.hpp>
#include <godot_cpp/classes/menu_button.hpp>
#include <godot_cpp/classes/popup_menu.hpp>
#include <godot_cpp/classes/rich_text_label.hpp>
#include <godot_cpp/classes/shortcut.hpp>

#include <arviz_godot/lib/avi/gui/CameraModeMenu.h>
#include <arviz_godot/lib/avi/gui/SelectionMenu.h>
#include <arviz_godot/lib/avi/gui/StatusIcons.h>
#include <arviz_godot/lib/avi/core/Utilities.h>

void
avi::gui::UserInterface::_bind_methods()
{
    DETOOLIFY;
    godot::ClassDB::bind_method(godot::D_METHOD("_scene_menu_select"), &UserInterface::_scene_menu_select);
    godot::ClassDB::bind_method(godot::D_METHOD("_view_menu_select"), &UserInterface::_view_menu_select);
}

avi::gui::UserInterface::UserInterface() = default;

void
avi::gui::UserInterface::_ready()
{
    DETOOLIFY;

    layerSelection = get_node<avi::gui::LayerSelection>("Overlays/Layers");
    graphicsSettings = get_node<avi::gui::GraphicsSettings>("Overlays/Visuals");
    replayCockpit = get_node<avi::gui::ReplayCockpit>("Overlays/Replay");

    settingsWindow = get_node<avi::gui::SettingsWindow>("Settings");
    orientationWindow = get_node<avi::gui::OrientationWindow>("Orientation");

    get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/Scene Menu")
        ->get_popup()
        ->connect("id_pressed", godot::Callable(this, "_scene_menu_select"));

    get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/View Menu")
        ->get_popup()
        ->connect("id_pressed", godot::Callable(this, "_view_menu_select"));

    helpWindow = get_node<godot::Window>("Help");
}

void
avi::gui::UserInterface::setupUI(avi::core::EventInvokingSystem& eventInvokingSystem,
                                 avi::core::EventSubscriptionSystem& eventSubscriptionSystem,
                                 avi::core::EventToKeyProvider& eventToKeyProvider,
                                 avi::core::Settings& settings)
{
    DETOOLIFY;
    this->eventInvokingSystem = &eventInvokingSystem;
    this->eventSubscriptionSystem = &eventSubscriptionSystem;

    settingsWindow->setup(settings);

    setupConstantHelpBar(eventToKeyProvider);

    get_node<avi::gui::CameraModeMenu>("Bar/Bar Background/Bar Container/CameraMenu")
        ->setup(eventInvokingSystem, eventSubscriptionSystem, eventToKeyProvider);

    eventSubscriptionSystem.subscribe("HIDE_UI",
                                      [this]
                                      {
                                          if (hiddenToggle)
                                          {
                                              requestVisibleUI();
                                          }
                                          else
                                          {
                                              requestHiddenUI();
                                          }

                                          hiddenToggle = !hiddenToggle;
                                      });

    viewMenuOptions = {"VIEW_CAMERA_RESET",
                       "VIEW_CAMERA_FOCUS_SCENE",
                       "TAKE_SCREENSHOT",
                       "RECORD_VIDEO",
                       "HIDE_UI",
                       "VIEW_CAMERA_SWITCH_PROJECTION",
                       "PLACE_ON_X_NUMPAD",
                       "PLACE_ON_Y_NUMPAD",
                       "PLACE_ON_Z_NUMPAD"};

    auto* viewMenu = get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/View Menu");
    viewMenu->set_disable_shortcuts(true);

    for (int index = 0; index < viewMenu->get_popup()->get_item_count(); index++)
    {
        std::string const& event = viewMenuOptions[index];
        std::optional<godot::InputEventKey*> key = eventToKeyProvider.getEventKey(event);

        if (key.has_value())
        {
            godot::Ref<godot::Shortcut> shortCut = memnew(godot::Shortcut);

            godot::Array events;
            events.push_back(key.value());
            shortCut->set_events(events);

            viewMenu->get_popup()->set_item_shortcut(index, shortCut);
        }
    }


    sceneMenuOptions = {"OPEN_SCREENSHOT_DIRECTORY",
                        "OPEN_VIDEO_DIRECTORY",
                        "ORIENTATION",
                        "SETTINGS",
                        "HELP"};

    auto* sceneMenu = get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/Scene Menu");
    sceneMenu->set_disable_shortcuts(true);

    for (int index = 0; index < sceneMenu->get_popup()->get_item_count(); index++)
    {
        std::string const& event = sceneMenuOptions[index];
        std::optional<godot::InputEventKey*> key = eventToKeyProvider.getEventKey(event);

        if (key.has_value())
        {
            godot::Ref<godot::Shortcut> shortCut = memnew(godot::Shortcut);

            godot::Array events;
            events.push_back(key.value());
            shortCut->set_events(events);

            sceneMenu->get_popup()->set_item_shortcut(index, shortCut);
        }
    }

    eventSubscriptionSystem.subscribe(sceneMenuOptions[0],[this]() -> void {_scene_menu_select(0);});
    eventSubscriptionSystem.subscribe(sceneMenuOptions[1],[this]() -> void {_scene_menu_select(1);});
    eventSubscriptionSystem.subscribe(sceneMenuOptions[2],[this]() -> void {_scene_menu_select(2);});
    eventSubscriptionSystem.subscribe(sceneMenuOptions[3],[this]() -> void {_scene_menu_select(3);});
    eventSubscriptionSystem.subscribe(sceneMenuOptions[4],[this]() -> void {_scene_menu_select(4);});
}

void
avi::gui::UserInterface::setupConstantHelpBar(avi::core::EventToKeyProvider& eventToKeyProvider)
{
    DETOOLIFY;
    auto* label = get_node<godot::RichTextLabel>("Bar/Bar Background/Bar Container/Constant Hint");

    std::string text = "[center]";

    auto addHint = [&text, &eventToKeyProvider](const char* event, const std::string& description)
    {
        std::optional<godot::InputEventKey*> key = eventToKeyProvider.getEventKey(event);

        std::string hint;

        if (key.has_value())
        {
            hint += " [";
            hint += avi::core::utilities::to_string(key.value()->as_text());
            hint += "] ";
        }

        hint += description;
        text += hint;
    };

    addHint("VIEW_CAMERA_FOCUS_POINT", "Focus");
    addHint("VIEW_CAMERA_FOCUS_SCENE", "View All");
    addHint("VIEW_CAMERA_RESET", "Reset View");
    addHint("HIDE_UI", "Toggle UI");
    addHint("TAKE_SCREENSHOT", "Capture Screenshot");
    addHint("RECORD_VIDEO", "Record Video");
    addHint("VIEW_CAMERA_LOCK", "Lock Mouse");

    text += "[/center]";

    label->set_use_bbcode(true);
    label->parse_bbcode(text.c_str());
}

void
avi::gui::UserInterface::setupStatusIcons(avi::connection::Connectable& connectable,
                                          avi::connection::StorageModeSwitch& mode,
                                          avi::recording::VideoRecorder& recorder,
                                          avi::scene::GraphicsController& graphicsController,
                                          avi::core::EventToKeyProvider& eventToKeyProvider)
{
    DETOOLIFY;
    this->connectable = &connectable;

    auto* statusIcons = get_node<avi::gui::StatusIcons>("StatusBar");
    statusIcons->setup(connectable, mode, recorder, graphicsController, eventToKeyProvider);

    advertiser = statusIcons;
}

void
avi::gui::UserInterface::setupScene(avi::scene::VisualizedScene& visualizedScene,
                                    avi::recording::VideoRecorder& videoRecorder,
                                    avi::recording::ImageRecorder& imageRecorder)
{
    DETOOLIFY;
    this->videoRecorder = &videoRecorder;
    this->imageRecorder = &imageRecorder;
    layerSelection->setup(visualizedScene);
    orientationWindow->setup(visualizedScene);
}

void
avi::gui::UserInterface::setupGraphics(avi::scene::GraphicsController& graphicsController,
                                       avi::scene::ProfileStorage& profileStorage)
{
    DETOOLIFY;
    graphicsSettings->setup(graphicsController, profileStorage);
}

void
avi::gui::UserInterface::setupInteraction(
    avi::interaction::InteractionDispatch& interactionDispatch)
{
    DETOOLIFY;
    this->interactionDispatch = &interactionDispatch;
    this->selectionMenu = get_node<avi::gui::SelectionMenu>("InteractionPopupMenu");
}

void
avi::gui::UserInterface::setupReplay(avi::replay::ReplayController& replayController,
                                     avi::scene::VisualizedScene& visualizedScene)
{
    DETOOLIFY;
    this->replayCockpit->setup(replayController, visualizedScene, this);
}

void
avi::gui::UserInterface::setupManipulation(avi::core::EventToKeyProvider& eventToKeyProvider)
{
    DETOOLIFY;
    manipulationMenu =
        get_node<avi::gui::ManipulationMenu>("Bar/Bar Background/Bar Container/ManipulationMenu");
    manipulationMenu->setup(
        *interactionDispatch->getManipulator(), *eventSubscriptionSystem, eventToKeyProvider);
}

void
avi::gui::UserInterface::_scene_menu_select(int64_t id)
{
    DETOOLIFY;

    helpWindow->hide();
    orientationWindow->hide();
    settingsWindow->hide();

    switch(id) {
        case 0: // Open Screenshot Directory
        {
            godot::PackedStringArray poolStringArray;
            poolStringArray.push_back(imageRecorder->getLastScreenshotPath().c_str());

            godot::OS::get_singleton()->execute("nautilus", poolStringArray);
            break;
        }
        case 1: // Open Video Directory
        {
            godot::PackedStringArray poolStringArray;
            poolStringArray.push_back(videoRecorder->getLastRecordingPath().c_str());

            godot::OS::get_singleton()->execute("nautilus", poolStringArray);
            break;
        }
        case 2: // Orientation
            orientationWindow->open();
            break;
        case 3: // Settings
            settingsWindow->open();
            break;
        case 4: // Help
            helpWindow->popup_centered();
            break;
    }
}

void
avi::gui::UserInterface::_view_menu_select(int64_t id)
{
    DETOOLIFY;
    if (id >= 0 && id < static_cast<int64_t>(viewMenuOptions.size()))
    {
        eventInvokingSystem->invoke(viewMenuOptions[id]);
    }
    else
    {
        ARMARX_ERROR << "Unsupported option in menu 'view'";
    }
}

void
avi::gui::UserInterface::requestHiddenUI()
{
    DETOOLIFY;
    hidingCounter++;

    if (hidingCounter > 0)
    {
        set_visible(false);
    }
}

void
avi::gui::UserInterface::requestVisibleUI()
{
    DETOOLIFY;
    assert(hidingCounter > 0 && "More visible requests than hidden requests.");

    hidingCounter--;

    if (hidingCounter == 0)
    {
        set_visible(true);
    }
}

void
avi::gui::UserInterface::openSelectionMenu(godot::Vector2 mousePosition,
                                           std::vector<std::string> interactions,
                                           std::function<void(int64_t)> callback)
{
    DETOOLIFY;
    this->selectionMenu->open(mousePosition, interactions, callback);
}

void
avi::gui::UserInterface::openTextInputPrompt()
{
    DETOOLIFY;
}

void
avi::gui::UserInterface::clear()
{
    DETOOLIFY;
    layerSelection->clear();
}

void
avi::gui::UserInterface::lockLayerSelection()
{
    DETOOLIFY;
    layerSelection->lock();
}

void
avi::gui::UserInterface::unlockLayerSelection()
{
    DETOOLIFY;
    layerSelection->unlock();
}

avi::core::ControlAdvertiser&
avi::gui::UserInterface::getAdvertiser()
{
    return *advertiser;
}

