#pragma once

#include <filesystem>
#include <functional>
#include <vector>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/check_button.hpp>
#include <godot_cpp/classes/color_picker_button.hpp>
#include <godot_cpp/classes/label.hpp>
#include <godot_cpp/classes/line_edit.hpp>
#include <godot_cpp/classes/option_button.hpp>
#include <godot_cpp/classes/packed_scene.hpp>
#include <godot_cpp/classes/resource_loader.hpp>
#include <godot_cpp/classes/spin_box.hpp>
#include <godot_cpp/classes/viewport.hpp>
#include <godot_cpp/classes/control.hpp>

#include <arviz_godot/lib/avi/core/ValueOption.h>
#include <arviz_godot/lib/avi/scene/DrawStyle.h>
#include <arviz_godot/lib/avi/core/Utilities.h>

namespace avi::gui
{
    class VisualOptionTransformer
    {
    private:
        template <typename C>
        static C*
        create(godot::Control* root,
               const std::string& resource,
               const std::string& label,
               const std::string& node,
               const std::string& tooltip)
        {
            godot::ResourceLoader* loader = godot::ResourceLoader::get_singleton();
            godot::Ref<godot::PackedScene> element = loader->load(resource.c_str());

            auto* elementRoot = godot::Object::cast_to<godot::Control>(element->instantiate());

            elementRoot->set_name(label.c_str());

            auto* labelNode = elementRoot->get_node<godot::Label>("Label");
            labelNode->set_text(label.c_str());

            if (!tooltip.empty()) {
                labelNode->set_tooltip_text(tooltip.c_str());
            }

            root->add_child(elementRoot);

            return elementRoot->get_node<C>(node.c_str());
        }

    public:
        std::vector<std::function<void()>>
        createGraphicsControls(godot::Control* root,
                               godot::Node* receiver,
                               std::vector<avi::core::Option<VisualOptionTransformer>*>&& options);

        template <typename V>
        void transform(avi::core::ValueOption<V, VisualOptionTransformer>& option);

        template <typename V, typename D = void>
        struct Transform
        {
            static void
            transform(avi::core::ValueOption<V, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
            }
        };

        template <typename D>
        struct Transform<godot::Color, D>
        {
            static void
            transform(avi::core::ValueOption<godot::Color, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::ColorPickerButton>(
                    root, "res://scenes/options/color.tscn", option.getLabel(), "Color", option.getToolTip());
                values->connect("color_changed", godot::Callable(receiver, "_change"));

                values->set_pick_color(option.getValue());

                std::function<void()> fn = [values, &option]()
                { option.setValue(values->get_pick_color()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<double, D>
        {
            static void
            transform(avi::core::ValueOption<double, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::SpinBox>(
                    root, "res://scenes/options/double.tscn", option.getLabel(), "Double", option.getToolTip());
                values->connect("value_changed", godot::Callable(receiver, "_change"));

                values->set_value(static_cast<float>(option.getValue()));

                std::function<void()> fn = [values, &option]()
                { option.setValue(values->get_value()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<int64_t, D>
        {
            static void
            transform(avi::core::ValueOption<int64_t, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::SpinBox>(
                    root, "res://scenes/options/integer.tscn", option.getLabel(), "Integer", option.getToolTip());
                values->connect("value_changed", godot::Callable(receiver, "_change"));

                int64_t value = option.getValue();
                values->set_value(static_cast<float>(value));

                std::function<void()> fn = [values, &option]()
                {
                    float newValue = values->get_value();
                    option.setValue(static_cast<int64_t>(newValue));
                };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<bool, D>
        {
            static void
            transform(avi::core::ValueOption<bool, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* value = create<godot::CheckButton>(
                    root, "res://scenes/options/boolean.tscn", option.getLabel(), "Boolean", option.getToolTip());
                value->connect("toggled", godot::Callable(receiver, "_change"));

                value->set_pressed(option.getValue());

                std::function<void()> fn = [value, &option]()
                { option.setValue(value->is_pressed()); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<godot::Viewport::MSAA, D>
        {
            static void
            transform(
                avi::core::ValueOption<godot::Viewport::MSAA, VisualOptionTransformer>& option,
                godot::Control* root,
                godot::Node* receiver,
                std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::OptionButton>(
                    root, "res://scenes/options/option.tscn", option.getLabel(), "Option", option.getToolTip());
                values->connect("item_selected", godot::Callable(receiver, "_change"));

                values->add_item("Disabled");
                values->add_item("2x");
                values->add_item("4x");
                values->add_item("8x");

                int32_t select;

                switch (option.getValue())
                {
                    case godot::Viewport::MSAA_DISABLED:
                        select = 0;
                        break;
                    case godot::Viewport::MSAA_2X:
                        select = 1;
                        break;
                    case godot::Viewport::MSAA_4X:
                        select = 2;
                        break;
                    case godot::Viewport::MSAA_8X:
                        select = 3;
                        break;
                    default:
                        select = 0;
                        break;
                }

                values->select(select);

                std::function<void()> fn = [values, &option]()
                {
                    auto idx = values->get_selected();
                    switch (idx)
                    {
                        case 0:
                            option.setValue(godot::Viewport::MSAA::MSAA_DISABLED);
                            break;
                        case 1:
                            option.setValue(godot::Viewport::MSAA::MSAA_2X);
                            break;
                        case 2:
                            option.setValue(godot::Viewport::MSAA::MSAA_4X);
                            break;
                        case 3:
                            option.setValue(godot::Viewport::MSAA::MSAA_8X);
                            break;

                        default:
                            option.setValue(godot::Viewport::MSAA::MSAA_DISABLED);
                            break;
                    }
                };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<avi::scene::DrawStyle, D>
        {
            static void
            transform(
                avi::core::ValueOption<avi::scene::DrawStyle, VisualOptionTransformer>& option,
                godot::Control* root,
                godot::Node* receiver,
                std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::OptionButton>(
                    root, "res://scenes/options/option.tscn", option.getLabel(), "Option", option.getToolTip());
                values->connect("item_selected", godot::Callable(receiver, "_change"));

                values->add_item("Original");
                values->add_item("Visualization");
                values->add_item("Collision");

                int64_t select = 0;

                switch (option.getValue())
                {
                    case avi::scene::DrawStyle::ORIGINAL:
                        select = 0;
                        break;
                    case avi::scene::DrawStyle::VISUALIZATION:
                        select = 1;
                        break;
                    case avi::scene::DrawStyle::COLLISION:
                        select = 2;
                        break;
                }

                values->select(select);

                std::function<void()> fn = [values, &option]()
                {
                    auto idx = values->get_selected();
                    switch (idx)
                    {
                        case 0:
                            option.setValue(avi::scene::DrawStyle::ORIGINAL);
                            break;
                        case 1:
                            option.setValue(avi::scene::DrawStyle::VISUALIZATION);
                            break;
                        case 2:
                            option.setValue(avi::scene::DrawStyle::COLLISION);
                            break;

                        default:
                            break;
                    }
                };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<std::string, D>
        {
            static void
            transform(avi::core::ValueOption<std::string, VisualOptionTransformer>& option,
                      godot::Control* root,
                      godot::Node* receiver,
                      std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::LineEdit>(
                    root, "res://scenes/options/string.tscn", option.getLabel(), "String", option.getToolTip());
                values->connect("text_changed", godot::Callable(receiver, "_change"));

                values->set_text(option.getValue().c_str());

                std::function<void()> fn = [values, &option]()
                {  option.setValue(avi::core::utilities::to_string(values->get_text())); };

                fns.push_back(fn);
            }
        };

        template <typename D>
        struct Transform<std::filesystem::path, D>
        {
            static void
            transform(
                avi::core::ValueOption<std::filesystem::path, VisualOptionTransformer>& option,
                godot::Control* root,
                godot::Node* receiver,
                std::vector<std::function<void()>>& fns)
            {
                auto* values = create<godot::LineEdit>(
                    root, "res://scenes/options/path.tscn", option.getLabel(), "Path", option.getToolTip());
                values->connect("text_changed", godot::Callable(receiver, "_change"));

                values->set_text(option.getValue().c_str());

                std::function<void()> fn = [values, &option]()
                { option.setValue(avi::core::utilities::to_string(values->get_text())); };

                fns.push_back(fn);
            }
        };

    private:
        godot::Control* currentRoot{nullptr};
        godot::Node* currentReceiver{nullptr};
        std::vector<std::function<void()>> returnFunctions;
    };

    template <typename V>
    void
    VisualOptionTransformer::transform(core::ValueOption<V, VisualOptionTransformer>& option)
    {
        Transform<V>::transform(option, currentRoot, currentReceiver, returnFunctions);
    }
} // namespace avi::gui
