#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "InteractionDispatch.h"

#include <cassert>
#include <iostream>

#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/input.hpp>
#include <godot_cpp/classes/input_event_mouse_button.hpp>
#include <godot_cpp/classes/control.hpp>


avi::interaction::InteractionDispatch::InteractionDispatch() {
    this->manipulator = std::make_unique<Manipulator>(*this);
}

void avi::interaction::InteractionDispatch::_bind_methods() {
    godot::ClassDB::bind_method(godot::D_METHOD("_notification"), &InteractionDispatch::_notification);
}

void
avi::interaction::InteractionDispatch::_notification(int what)
{
    DETOOLIFY;
    if (what == godot::Object::NOTIFICATION_PREDELETE)
    {
        manipulator->preDelete();
    }
}

void
avi::interaction::InteractionDispatch::setup(avi::connection::InteractionSink& interactionSink,
                                             avi::scene::CameraSet& cameraSet,
                                             avi::core::InterfaceUtility& userInterface,
                                             avi::connection::StorageModeSwitch& storageMode)
{
    DETOOLIFY;
    this->interactionSink = &interactionSink;
    this->cameraSet = &cameraSet;
    this->userInterface = &userInterface;
    this->storageMode = &storageMode;

    this->rootControl = get_node<godot::Control>("/root/Root/Control");

    avi::interaction::Move* move =
        get_node<avi::interaction::Move>("/root/Root/Control/3D/3D/Visualized Scene/Move");
    avi::interaction::Rotate* rotate =
        get_node<avi::interaction::Rotate>("/root/Root/Control/3D/3D/Visualized Scene/Rotate");
    avi::interaction::Scale* scale =
        get_node<avi::interaction::Scale>("/root/Root/Control/3D/3D/Visualized Scene/Scale");
    avi::interaction::None* none =
        get_node<avi::interaction::None>("/root/Root/Control/3D/3D/Visualized Scene/None");
    this->manipulator->setGizmos(*move, *rotate, *scale, *none);
}

void
avi::interaction::InteractionDispatch::sendInteraction(avi::connection::Interaction& interaction)
{
    DETOOLIFY;
    interactionSink->addInteraction(interaction);
}

void
avi::interaction::InteractionDispatch::sendManipulation(avi::connection::Interaction& interaction)
{
    DETOOLIFY;
    interactionSink->addInteraction(interaction);
}

avi::interaction::Manipulator*
avi::interaction::InteractionDispatch::getManipulator()
{
    return this->manipulator.get();
}

void
avi::interaction::InteractionDispatch::_process(double delta)
{
    DETOOLIFY;

    activeCamera = cameraSet->getActiveCamera();
    godot::Vector2 mousePosition = get_viewport()->get_mouse_position();
    auto target = activeCamera->getSelectedElement(mousePosition);

    bool hasHitSelectable = false;

    if (target.has_value() && target->first)
    {
        avi::core::Element* element = target->first;
        hasHitSelectable = element->isSelectable() || element->hasContextMenu();
    }

    godot::Input* input = godot::Input::get_singleton();

    if (hasHitSelectable)
    {
        input->set_default_cursor_shape(godot::Input::CURSOR_POINTING_HAND);
        rootControl->set_default_cursor_shape(godot::Control::CursorShape::CURSOR_POINTING_HAND);
    }
    else
    {
        input->set_default_cursor_shape(godot::Input::CURSOR_ARROW);
        rootControl->set_default_cursor_shape(godot::Control::CursorShape::CURSOR_ARROW);
    }
}

void
avi::interaction::InteractionDispatch::_unhandled_input(const godot::Ref<godot::InputEvent>& event)
{
    DETOOLIFY;
    if (storageMode->isLive())
    {
        auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
        if (mouseButtonEvent)
        {
            auto idx = mouseButtonEvent->get_button_index();
            assert(this->cameraSet != nullptr);
            activeCamera = cameraSet->getActiveCamera();
            assert(this->activeCamera != nullptr);


            if (idx == godot::MouseButton::MOUSE_BUTTON_RIGHT && event->is_pressed())
            {
                startInteraction();
            }

            if (idx == godot::MouseButton::MOUSE_BUTTON_LEFT && event->is_pressed())
            {
                startManipulation();
            }
        }
    }
}

void
avi::interaction::InteractionDispatch::startInteraction()
{
    DETOOLIFY;
    if (previousElement != nullptr)
    {
        deselect();
    }

    godot::Vector2 mousePosition = get_viewport()->get_mouse_position();
    this->targetElement = activeCamera->getSelectedElement(mousePosition);

    if (!targetElement.has_value())
    {
        return;
    }

    if (targetElement->first == nullptr || targetElement->second == nullptr)
    {
        return;
    }

    targetElement->first->setDeselectCallback([this]() { this->deselect(); });

    if (targetElement->first->hasContextMenu() && targetElement->first->getLayer()->isVisible())
    {
        enableManipulation = false;
        wasManipulation = false;
        std::vector<std::string> interactions = targetElement->first->getInteractions();
        userInterface->openSelectionMenu(mousePosition,
                                         interactions,
                                         [this](int64_t i)
                                         {
                                             this->interaction =
                                                 avi::connection::Interaction::createOption(
                                                     targetElement->first, i);
                                             sendInteraction(interaction);
                                         });
    }
}

void
avi::interaction::InteractionDispatch::startManipulation()
{
    DETOOLIFY;
    if (!enableManipulation)
    {
        enableManipulation = true;
        return;
    }

    godot::Vector2 mousePosition = get_viewport()->get_mouse_position();
    bool anyHit = false;
    this->targetElement = activeCamera->getSelectedElement(mousePosition, anyHit);

    if (!targetElement.has_value())
    {
        if (!anyHit)
            deselect();
        return;
    }

    if (targetElement->first->isSelectable() && targetElement->first->getLayer()->isVisible())
    {
        deselect();
        this->interaction = avi::connection::Interaction::createSelection(targetElement->first);
        sendInteraction(interaction);
        previousElement = targetElement->first;
        if (targetElement->first->isManipulable())
        {
            this->manipulator->target(
                dynamic_cast<avi::scene::ElementNode&>(*targetElement->first));
            wasManipulation = true;
        }

        previousElement->setDeselectCallback([this]() { this->deselect(); });
    }
    else
    {
        deselect();
    }
}

void
avi::interaction::InteractionDispatch::deselect()
{
    DETOOLIFY;

    if (previousElement != nullptr)
    {
        avi::connection::Interaction interaction =
            avi::connection::Interaction::createDeselection(previousElement);

        dynamic_cast<avi::scene::ElementNode*>(previousElement)->deselect();

        sendInteraction(interaction);
        previousElement = nullptr;
    }

    if (wasManipulation)
    {
        this->manipulator->untarget();
        wasManipulation = false;
    }
}

void
avi::interaction::InteractionDispatch::setupControlAdvertisement(
    avi::core::ControlAdvertiser& advertiser)
{
    DETOOLIFY;
    manipulator->setupControlAdvertisement(advertiser);
}

void
avi::interaction::InteractionDispatch::quit()
{
    deselect();
}
