/*
 * Adapted from: godot/editor/plugins/node_3d_editor_plugin.cpp
 */
#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "MoveAxisGizmo.h"

#include <sstream>

#include <godot_cpp/classes/input_event_mouse_button.hpp>
#include <godot_cpp/classes/viewport.hpp>
#include <godot_cpp/classes/world3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/interaction/Move.h>
#include <arviz_godot/lib/avi/interaction/godot/Manipulator.h>

avi::interaction::MoveAxisGizmo::MoveAxisGizmo()
{
    DETOOLIFY;
    this->set_capture_input_on_drag(true);
}

avi::interaction::MoveAxisGizmo::~MoveAxisGizmo() = default;

void
avi::interaction::MoveAxisGizmo::_bind_methods()
{
    DETOOLIFY;
}

void
avi::interaction::MoveAxisGizmo::_input(godot::Ref<godot::InputEvent> const& event)
{
    DETOOLIFY;
    if (auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr()))
    {
        auto idx = mouseButtonEvent->get_button_index();
        if (idx == 1 and not(mouseButtonEvent->is_pressed()))
        {
            if (moved)
            {
                move->updateRelease();
            }
            moved = false;
        }
    }
}

void
avi::interaction::MoveAxisGizmo::_input_event(godot::Camera3D* camera,
                                              const godot::Ref<godot::InputEvent>& event,
                                              const godot::Vector3& position,
                                              const godot::Vector3& normal,
                                              int32_t shape_idx)
{
    this->camera = camera;

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent and mouseButtonEvent->get_button_index() == 1)
    {
        bool pressed = mouseButtonEvent->is_pressed();

        if (moved and not pressed)
        {
            move->updateRelease();
        }
        else if (not moved and pressed)
        {
            data = godot::Interaction::compute(*mouseButtonEvent,
                                               *this->move,
                                               *this->move->get_proxy(),
                                               *this->camera,
                                               this->axis);
        }
        moved = pressed;
    }
}

void
avi::interaction::MoveAxisGizmo::_process(double)
{
    DETOOLIFY;
    if (not moved)
    {
        return;
    }

    godot::Vector2 mouse_pos = get_viewport()->get_mouse_position();
    godot::Vector3 mouse_ray_normal_global =
        godot::Interaction::get_ray_normal(*camera, mouse_pos).normalized();
    godot::Vector3 mouse_ray_origin_global = godot::Interaction::get_ray_origin(*camera, mouse_pos);

    auto update = [&](godot::Interaction& _edit)
    {
        using namespace godot;

        auto _get_global_camera_normal = [this]() -> Vector3
        { return -this->camera->get_global_transform().basis[2]; };

        auto get_global_gizmo_transform = [&]() -> Transform3D { return _edit.original_global; };

        Vector3 motion_mask;
        Plane plane_global;

        switch (_edit.plane)
        {
            case TRANSFORM_X_AXIS:
                motion_mask = get_global_gizmo_transform().basis[0];
                plane_global = Plane(
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized(),
                    _edit.center_global);
                break;
            case TRANSFORM_Y_AXIS:
                motion_mask = get_global_gizmo_transform().basis[1];
                plane_global = Plane(
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized(),
                    _edit.center_global);
                break;
            case TRANSFORM_Z_AXIS:
                motion_mask = get_global_gizmo_transform().basis[2];
                plane_global = Plane(
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized(),
                    _edit.center_global);
                break;
        }

        Vector3 intersection_global;
        if (not plane_global.intersects_ray(
                mouse_ray_origin_global, mouse_ray_normal_global, &intersection_global))
        {
            return;
        }

        Vector3 click_global;
        if (not plane_global.intersects_ray(
                _edit.click_ray_origin_global, _edit.click_ray_normal_global, &click_global))
        {
            return;
        }

        Vector3 motion_global = intersection_global - click_global;
        motion_global = motion_mask.dot(motion_global) * motion_mask;

        std::optional<float> snapPrecision =
            getPrecision(1000.0f / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR,
                         250.0f / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR,
                         10.0f / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

        if (snapPrecision.has_value())
        {
            motion_global.snap(godot::Vector3(1,1,1) * snapPrecision.value());
        }

        Transform3D new_global = _edit.proxy_original_global;
        new_global.origin += motion_global;

        move->updateTransform(new_global);
    };

    update(*this->data);
}

void
avi::interaction::MoveAxisGizmo::setup(avi::interaction::Move& move, int axis)
{
    DETOOLIFY;
    this->move = &move;
    this->axis = axis;

    if (not manipulator)
    {
        manipulator = std::make_unique<godot::Manipulator>();
        godot::Manipulator::AxisData axis_data = manipulator->make_axis_data(axis);
        manipulator->init_move_gizmo(axis_data);

        godot::MeshInstance3D* mesh = memnew(godot::MeshInstance3D);
        mesh->set_mesh(manipulator->move_gizmo[axis]);
        this->add_child(mesh);
    }
}

void
avi::interaction::MoveAxisGizmo::setActive(bool activeStatus)
{
    DETOOLIFY;
    if (activeStatus)
    {
        this->show();
    }
    else
    {
        this->hide();
    }
}
