/*
* Adapted from: godot/editor/plugins/node_3d_editor_plugin.cpp
*/

#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include <godot_cpp/classes/input_event_mouse_button.hpp>
#include <godot_cpp/classes/viewport.hpp>

#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/interaction/Rotate.h>
#include <arviz_godot/lib/avi/interaction/RotateAxisGizmo.h>
#include <arviz_godot/lib/avi/interaction/godot/Manipulator.h>

void
avi::interaction::RotateAxisGizmo::_bind_methods()
{
    DETOOLIFY;
}

avi::interaction::RotateAxisGizmo::RotateAxisGizmo()
{
    DETOOLIFY;
    this->set_capture_input_on_drag(true);
}

avi::interaction::RotateAxisGizmo::~RotateAxisGizmo()
{
    DETOOLIFY;
}

void
avi::interaction::RotateAxisGizmo::setup(avi::interaction::Rotate& rotate, int axis)
{
    DETOOLIFY;
    this->rotate = &rotate;
    this->axis = axis;

    if (not manipulator)
    {
        manipulator = std::make_unique<godot::Manipulator>();
        godot::Manipulator::AxisData axis_data = manipulator->make_axis_data(axis);
        manipulator->init_rotate_gizmo(axis_data);

        godot::MeshInstance3D* mesh = memnew(godot::MeshInstance3D);
        mesh->set_mesh(manipulator->rotate_gizmo[axis]);
        this->add_child(mesh);
    }
}

void
avi::interaction::RotateAxisGizmo::_input(godot::Ref<godot::InputEvent> const& event)
{
    DETOOLIFY;
    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent)
    {
        auto idx = mouseButtonEvent->get_button_index();
        if (idx == 1 && !(mouseButtonEvent->is_pressed()))
        {
            if (moved)
                rotate->updateRelease();
            moved = false;
        }
    }
}

void
avi::interaction::RotateAxisGizmo::_input_event(godot::Camera3D* camera,
                                                const godot::Ref<godot::InputEvent>& event,
                                                const godot::Vector3& position,
                                                const godot::Vector3& normal,
                                                int32_t shape_idx)
{
    this->camera = camera;

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent and mouseButtonEvent->get_button_index() == 1)
    {
        bool pressed = mouseButtonEvent->is_pressed();

        if (moved and not pressed)
        {
            rotate->updateRelease();
        }
        else if (not moved and pressed)
        {
            data = godot::Interaction::compute(*mouseButtonEvent,
                                               *this->rotate,
                                               *this->rotate->get_proxy(),
                                               *this->camera,
                                               (axis + 1) % 3);
        }

        moved = pressed;
    }
}

void
avi::interaction::RotateAxisGizmo::_process(double delta)
{
    DETOOLIFY;
    if (!moved)
        return;

    godot::Vector2 mouse_pos = get_viewport()->get_mouse_position();
    godot::Vector3 mouse_ray_normal_global =
        godot::Interaction::get_ray_normal(*camera, mouse_pos).normalized();
    godot::Vector3 mouse_ray_origin_global = godot::Interaction::get_ray_origin(*camera, mouse_pos);

    auto update = [&](godot::Interaction& _edit)
    {
        using namespace godot;

        auto _get_global_camera_normal = [this]() -> Vector3
        { return -this->camera->get_global_transform().basis[2]; };

        auto get_global_gizmo_transform = [&]() -> Transform3D { return _edit.original_global; };

        Plane plane = Plane(_get_global_camera_normal(), _edit.center_global);

        Vector3 local_axis;
        Vector3 global_axis;
        switch (_edit.plane)
        {
            case TRANSFORM_X_AXIS:
                local_axis = Vector3(1, 0, 0);
                break;
            case TRANSFORM_Y_AXIS:
                local_axis = Vector3(0, 1, 0);
                break;
            case TRANSFORM_Z_AXIS:
                local_axis = Vector3(0, 0, 1);
                break;
        }

        global_axis = get_global_gizmo_transform().basis.xform(local_axis).normalized();

        Vector3 intersection;
        if (!plane.intersects_ray(mouse_ray_origin_global, mouse_ray_normal_global, &intersection))
        {
            return;
        }

        Vector3 click;
        if (!plane.intersects_ray(
                _edit.click_ray_origin_global, _edit.click_ray_normal_global, &click))
        {
            return;
        }

        static const float orthogonal_threshold = Math::cos(Math::deg_to_rad(87.0f));
        bool axis_is_orthogonal = abs(plane.normal.dot(global_axis)) < orthogonal_threshold;

        double angle = 0.0f;
        if (axis_is_orthogonal)
        {
            Vector3 projection_axis = plane.normal.cross(global_axis);
            Vector3 delta = intersection - click;
            float projection = delta.dot(projection_axis);
            angle = (projection * (Math_PI / 2.0f)) / (1.0f * 1.1f);
        }
        else
        {
            Vector3 click_axis = (click - _edit.center_global).normalized();
            Vector3 current_axis = (intersection - _edit.center_global).normalized();
            angle = avi::core::utilities::signed_angle(click_axis, current_axis, global_axis);
        }

        std::optional<float> snapPrecision = getPrecision(30.0f, 5.0f, 1.0f);

        if (snapPrecision.has_value())
        {
            double snap = snapPrecision.value();

            angle = godot::Math::rad_to_deg(angle) + snap * 0.5;
            angle -= godot::Math::fmod(angle, snap);
            angle = godot::Math::deg_to_rad(angle);
        }

        auto compute_rotation = [&](const godot::Transform3D& original,
                                    const godot::Transform3D& original_local,
                                    bool is_local) -> Transform3D
        {
            Transform3D system = _edit.original_local;
            Vector3 rotation_axis = system.basis.xform(local_axis);

            Transform3D r;

            r.basis = Basis(rotation_axis.normalized(), angle) * original_local.basis;
            r.origin = original_local.origin;

            return r;
        };

        Transform3D rotation =
            compute_rotation(_edit.proxy_original_global, _edit.proxy_original_local, false);

        rotate->updateTransform(rotation);
    };

    update(*this->data);
}

void
avi::interaction::RotateAxisGizmo::setActive(bool activeStatus)
{
    DETOOLIFY;
    if (activeStatus)
    {
        this->show();
    }
    else
    {
        this->hide();
    }
}
