/*
* Adapted from: godot/editor/plugins/node_3d_editor_plugin.cpp
*/

#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ScaleAxisGizmo.h"

#include <memory>

#include <godot_cpp/classes/input_event_mouse_button.hpp>
//#include <godot_cpp/classes/math.hpp>
#include <godot_cpp/classes/viewport.hpp>


#include <godot_cpp/variant/vector3.hpp>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/interaction/Scale.h>
#include <arviz_godot/lib/avi/interaction/godot/Manipulator.h>

void
avi::interaction::ScaleAxisGizmo::_bind_methods()
{
    DETOOLIFY;
}

avi::interaction::ScaleAxisGizmo::ScaleAxisGizmo()
{
    DETOOLIFY;
    this->set_capture_input_on_drag(true);
}

avi::interaction::ScaleAxisGizmo::~ScaleAxisGizmo() = default;

void
avi::interaction::ScaleAxisGizmo::_input(godot::Ref<godot::InputEvent> const& event)
{
    DETOOLIFY;
    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent)
    {
        auto idx = mouseButtonEvent->get_button_index();
        if (idx == 1 && !(mouseButtonEvent->is_pressed()))
        {
            if (moved)
                scale->updateRelease();
            moved = false;
        }
    }
}

void
avi::interaction::ScaleAxisGizmo::_input_event(godot::Camera3D* camera,
                                               const godot::Ref<godot::InputEvent>& event,
                                               const godot::Vector3& position,
                                               const godot::Vector3& normal,
                                               int32_t shape_idx)
{
    this->camera = camera;

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent and mouseButtonEvent->get_button_index() == 1)
    {
        bool pressed = mouseButtonEvent->is_pressed();

        if (moved and not pressed)
        {
            scale->updateRelease();
        }
        else if (not moved and pressed)
        {
            data = godot::Interaction::compute(*mouseButtonEvent,
                                               *this->scale,
                                               *this->scale->get_proxy(),
                                               *this->camera,
                                               this->axis);
        }
        moved = pressed;
    }
}

void
avi::interaction::ScaleAxisGizmo::_process(double delta)
{
    DETOOLIFY;
    if (not moved)
        return;

    godot::Vector2 mouse_pos = get_viewport()->get_mouse_position();
    godot::Vector3 mouse_ray_normal_global =
        godot::Interaction::get_ray_normal(*camera, mouse_pos).normalized();
    godot::Vector3 mouse_ray_origin_global = godot::Interaction::get_ray_origin(*camera, mouse_pos);

    auto update = [&](godot::Interaction& _edit)
    {
        using namespace godot;

        auto _get_global_camera_normal = [this]() -> Vector3
        { return -this->camera->get_global_transform().basis[2]; };

        auto get_global_gizmo_transform = [&]() -> Transform3D { return _edit.original_global; };

        Vector3 motion_mask;
        Plane plane;

        switch (_edit.plane)
        {
            case TRANSFORM_X_AXIS:
                motion_mask = get_global_gizmo_transform().basis[0].normalized();
                plane = Plane(
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized(),
                    _edit.center_global);
                break;
            case TRANSFORM_Y_AXIS:
                motion_mask = get_global_gizmo_transform().basis[1].normalized();
                plane = Plane(
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized(),
                    _edit.center_global);
                break;
            case TRANSFORM_Z_AXIS:
                motion_mask = get_global_gizmo_transform().basis[2].normalized();
                plane = Plane(
                    motion_mask.cross(motion_mask.cross(_get_global_camera_normal())).normalized(),
                    _edit.center_global);
                break;
        }

        Vector3 intersection;
        if (!plane.intersects_ray(mouse_ray_origin_global, mouse_ray_normal_global, &intersection))
        {
            return;
        }

        Vector3 click;
        if (!plane.intersects_ray(
                _edit.click_ray_origin_global, _edit.click_ray_normal_global, &click))
        {
            return;
        }

        Vector3 motion = intersection - click;
        motion = motion_mask.dot(motion) * motion_mask;

        motion /= click.distance_to(_edit.center_global);
        motion = _edit.original_global.basis.inverse().xform(motion);

        std::optional<float> snapPrecision = getPrecision(10.00f, 2.5f, 0.1f);

        if (snapPrecision.has_value())
        {
            motion.snap(godot::Vector3(1,1,1) * snapPrecision.value());
        }

        auto compute_transform = [&](const godot::Transform3D& original,
                                     const godot::Transform3D& original_local,
                                     bool is_local) -> Transform3D
        {
            Transform3D s;

            if (is_local)
            {
                s.basis = avi::core::utilities::scaled_local(original_local.basis,
                                                             motion + Vector3(1, 1, 1));
                s.origin = original_local.origin;
            }
            else
            {
                s.basis = original_local.basis.scaled(motion + Vector3(1, 1, 1));
                s.origin = original_local.origin;
            }

            return s;
        };

        Transform3D new_global =
            compute_transform(_edit.proxy_original_global, _edit.proxy_original_local, false);
        Transform3D new_local =
            compute_transform(_edit.proxy_original_global, _edit.proxy_original_local, true);

        scale->updateTransform(new_global, new_local);
    };

    update(*this->data);
}

void
avi::interaction::ScaleAxisGizmo::setup(avi::interaction::Scale& scale, int axis)
{
    DETOOLIFY;
    this->scale = &scale;
    this->axis = axis;

    if (not manipulator)
    {
        manipulator = std::make_unique<godot::Manipulator>();
        godot::Manipulator::AxisData axis_data = manipulator->make_axis_data(axis);
        manipulator->init_scale_gizmo(axis_data);

        godot::MeshInstance3D* mesh = memnew(godot::MeshInstance3D);
        mesh->set_mesh(manipulator->scale_gizmo[axis]);
        this->add_child(mesh);
    }
}

void
avi::interaction::ScaleAxisGizmo::setActive(bool activeStatus)
{
    DETOOLIFY;
    if (activeStatus)
    {
        this->show();
    }
    else
    {
        this->hide();
    }
}
