/*
 * Adapted from Godot Engine
 */

#include "Manipulator.h"

#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/standard_material3d.hpp>

#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/resource_loader.hpp>
#include <godot_cpp/classes/shader.hpp>
#include <godot_cpp/classes/shader_material.hpp>
#include <godot_cpp/classes/surface_tool.hpp>

namespace godot
{

    void
    Manipulator::init_all_gizmos()
    {
        for (int axis = 0; axis < 3; axis++)
        {
            init_all_gizmos(axis);
        }
    }

    void
    Manipulator::init_all_gizmos(int axis)
    {
        const AxisData data = make_axis_data(axis);

        init_move_gizmo(data);
        init_move_plane_gizmo(data);
        init_rotate_gizmo(data);
        init_scale_gizmo(data);
        init_scale_plane_gizmo(data);
    }

    Manipulator::AxisData
    Manipulator::make_axis_data(int axis)
    {
        Color col = axis < 3 ? params.colors[axis] : Color();
        col.a = params.alpha;

        Ref<StandardMaterial3D> mat = memnew(StandardMaterial3D);
        mat->set_shading_mode(StandardMaterial3D::SHADING_MODE_UNSHADED);
        mat->set_transparency(StandardMaterial3D::TRANSPARENCY_DISABLED);
        mat->set_albedo(col);
        gizmo_color[axis] = mat;

        Ref<StandardMaterial3D> mat_hl = mat->duplicate();
        const Color albedo = col.from_hsv(col.get_h(), 0.25, 1.0, 1);
        mat_hl->set_albedo(albedo);
        gizmo_color_hl[axis] = mat_hl;

        Vector3 ivec;
        ivec[axis] = 1;

        Vector3 nivec;
        nivec[(axis + 1) % 3] = 1;
        nivec[(axis + 2) % 3] = 1;

        Vector3 ivec2;
        ivec2[(axis + 1) % 3] = 1;

        Vector3 ivec3;
        ivec3[(axis + 2) % 3] = 1;

        AxisData data;
        data.axis = axis;
        data.col = col;
        data.albedo = albedo;
        data.mat = mat;
        data.ivec = ivec;
        data.ivec2 = ivec2;
        data.ivec3 = ivec3;
        data.nivec = nivec;

        return data;
    }

    void
    Manipulator::init_move_gizmo(const AxisData& d)
    {
        // translate
        move_gizmo[d.axis] = godot::Ref<godot::ArrayMesh>(memnew(godot::ArrayMesh));

        Ref<SurfaceTool> surftool = memnew(SurfaceTool());
        surftool->begin(Mesh::PRIMITIVE_TRIANGLES);

        // Arrow profile
        const int arrow_points = 5;
        Vector3 arrow[5] = {
            d.nivec * 0.0 + d.ivec * 0.0,
            d.nivec * 0.02 + d.ivec * 0.0,
            d.nivec * 0.02 + d.ivec * params.gizmo_arrow_offset,
            d.nivec * 0.065 + d.ivec * params.gizmo_arrow_offset,
            d.nivec * 0.0 + d.ivec * (params.gizmo_arrow_offset + params.gizmo_arrow_size),
        };

        int arrow_sides = 16;

        for (int k = 0; k < arrow_sides; k++)
        {
            Basis ma(d.ivec, Math_PI * 2 * float(k) / arrow_sides);
            Basis mb(d.ivec, Math_PI * 2 * float(k + 1) / arrow_sides);

            for (int j = 0; j < arrow_points - 1; j++)
            {
                Vector3 points[4] = {
                    ma.xform(arrow[j]),
                    mb.xform(arrow[j]),
                    mb.xform(arrow[j + 1]),
                    ma.xform(arrow[j + 1]),
                };
                surftool->add_vertex(points[0]);
                surftool->add_vertex(points[1]);
                surftool->add_vertex(points[2]);

                surftool->add_vertex(points[0]);
                surftool->add_vertex(points[2]);
                surftool->add_vertex(points[3]);
            }
        }

        surftool->set_material(d.mat);
        surftool->commit(move_gizmo[d.axis]);
    }

    void
    Manipulator::init_move_plane_gizmo(const AxisData& d)
    {
        // Plane Translation
        move_plane_gizmo[d.axis] = Ref<ArrayMesh>(memnew(ArrayMesh()));

        Ref<SurfaceTool> surftool = memnew(SurfaceTool());
        surftool->begin(Mesh::PRIMITIVE_TRIANGLES);

        Vector3 vec = d.ivec2 - d.ivec3;
        Vector3 plane[4] = {vec * params.gizmo_plane_dst,
                            vec * params.gizmo_plane_dst + d.ivec2 * params.gizmo_plane_size,
                            vec * (params.gizmo_plane_dst + params.gizmo_plane_size),
                            vec * params.gizmo_plane_dst - d.ivec3 * params.gizmo_plane_size};

        Basis ma(d.ivec, Math_PI / 2);

        Vector3 points[4] = {
            ma.xform(plane[0]),
            ma.xform(plane[1]),
            ma.xform(plane[2]),
            ma.xform(plane[3]),
        };
        surftool->add_vertex(points[0]);
        surftool->add_vertex(points[1]);
        surftool->add_vertex(points[2]);

        surftool->add_vertex(points[0]);
        surftool->add_vertex(points[2]);
        surftool->add_vertex(points[3]);

        Ref<StandardMaterial3D> plane_mat = memnew(StandardMaterial3D());
        plane_mat->set_shading_mode(StandardMaterial3D::SHADING_MODE_UNSHADED);
        //plane_mat->set_on_top_of_alpha();
        plane_mat->set_transparency(StandardMaterial3D::TRANSPARENCY_DISABLED);
        plane_mat->set_cull_mode(StandardMaterial3D::CULL_DISABLED);
        plane_mat->set_albedo(d.col);
        plane_gizmo_color[d.axis] = plane_mat; // needed, so we can draw planes from both sides
        surftool->set_material(plane_mat);
        surftool->commit(move_plane_gizmo[d.axis]);

        Ref<StandardMaterial3D> plane_mat_hl = plane_mat->duplicate();
        plane_mat_hl->set_albedo(d.albedo);
        plane_gizmo_color_hl[d.axis] =
            plane_mat_hl; // needed, so we can draw planes from both sides
    }

    void
    Manipulator::init_rotate_gizmo(const AxisData& d)
    {
        // Rotate
        rotate_gizmo[d.axis] = Ref<ArrayMesh>(memnew(ArrayMesh()));

        Ref<SurfaceTool> surftool = memnew(SurfaceTool());
        surftool->begin(Mesh::PRIMITIVE_TRIANGLES);

        int n = 128; // number of circle segments
        int m = 30; // number of thickness segments

        for (int j = 0; j < n; ++j)
        {
            Basis basis = Basis(d.ivec, (Math_PI * 2.0f * j) / n);
            Vector3 vertex = basis.xform(d.ivec2 * params.gizmo_circle_size);
            for (int k = 0; k < m; ++k)
            {
                Vector2 ofs =
                    Vector2(Math::cos((Math_PI * 2.0 * k) / m), Math::sin((Math_PI * 2.0 * k) / m));
                Vector3 normal = d.ivec * ofs.x + d.ivec2 * ofs.y;
                surftool->set_normal(basis.xform(normal));
                surftool->add_vertex(vertex);
            }
        }

        for (int j = 0; j < n; ++j)
        {
            for (int k = 0; k < m; ++k)
            {
                int current_ring = j * m;
                int next_ring = ((j + 1) % n) * m;
                int current_segment = k;
                int next_segment = (k + 1) % m;

                surftool->add_index(current_ring + next_segment);
                surftool->add_index(current_ring + current_segment);
                surftool->add_index(next_ring + current_segment);

                surftool->add_index(next_ring + current_segment);
                surftool->add_index(next_ring + next_segment);
                surftool->add_index(current_ring + next_segment);
            }
        }

        Ref<Shader> rotate_shader = memnew(Shader());

        rotate_shader->set_code("\n"
                                "shader_type spatial; \n"
                                "render_mode unshaded, depth_test_disabled; \n"
                                "uniform vec4 albedo; \n"
                                "\n"
                                "mat3 orthonormalize(mat3 m) { \n"
                                "	vec3 x = normalize(m[0]); \n"
                                "	vec3 y = normalize(m[1] - x * dot(x, m[1])); \n"
                                "	vec3 z = m[2] - x * dot(x, m[2]); \n"
                                "	z = normalize(z - y * (dot(y,m[2]))); \n"
                                "	return mat3(x,y,z); \n"
                                "} \n"
                                "\n"
                                "void vertex() { \n"
                                "	mat3 mv = orthonormalize(mat3(MODELVIEW_MATRIX)); \n"
                                "	vec3 n = mv * VERTEX; \n"
                                "	float orientation = dot(vec3(0,0,-1),n); \n"
                                "	if (orientation <= 0.005) { \n"
                                "		VERTEX += NORMAL*0.02; \n"
                                "	} \n"
                                "} \n"
                                "\n"
                                "void fragment() { \n"
                                "	ALBEDO = albedo.rgb; \n"
                                "	ALPHA = 1.0; \n"
                                "}");

        Ref<ShaderMaterial> rotate_mat = memnew(ShaderMaterial());
        rotate_mat->set_render_priority(Material::RENDER_PRIORITY_MAX);
        rotate_mat->set_shader(rotate_shader);
        rotate_mat->set_shader_parameter("albedo", d.col);
        rotate_gizmo_color[d.axis] = rotate_mat;

        Array arrays = surftool->commit_to_arrays();
        rotate_gizmo[d.axis]->add_surface_from_arrays(Mesh::PRIMITIVE_TRIANGLES, arrays);
        rotate_gizmo[d.axis]->surface_set_material(0, rotate_mat);

        Ref<ShaderMaterial> rotate_mat_hl = rotate_mat->duplicate();
        rotate_mat_hl->set_shader_parameter("albedo", d.albedo);
        rotate_gizmo_color_hl[d.axis] = rotate_mat_hl;

        if (d.axis == 2)
        { // Rotation white outline
            Ref<ShaderMaterial> border_mat = rotate_mat->duplicate();

            Ref<Shader> border_shader = memnew(Shader());
            border_shader->set_code(
                "\n"
                "shader_type spatial; \n"
                "render_mode unshaded, depth_test_disabled; \n"
                "uniform vec4 albedo; \n"
                "\n"
                "mat3 orthonormalize(mat3 m) { \n"
                "	vec3 x = normalize(m[0]); \n"
                "	vec3 y = normalize(m[1] - x * dot(x, m[1])); \n"
                "	vec3 z = m[2] - x * dot(x, m[2]); \n"
                "	z = normalize(z - y * (dot(y,m[2]))); \n"
                "	return mat3(x,y,z); \n"
                "} \n"
                "\n"
                "void vertex() { \n"
                "	mat3 mv = orthonormalize(mat3(MODELVIEW_MATRIX)); \n"
                "	mv = inverse(mv); \n"
                "	VERTEX += NORMAL*0.008; \n"
                "	vec3 camera_dir_local = mv * vec3(0,0,1); \n"
                "	vec3 camera_up_local = mv * vec3(0,1,0); \n"
                "	mat3 rotation_matrix = mat3(cross(camera_dir_local, camera_up_local), "
                "camera_up_local, camera_dir_local); \n"
                "	VERTEX = rotation_matrix * VERTEX; \n"
                "} \n"
                "\n"
                "void fragment() { \n"
                "	ALBEDO = albedo.rgb; \n"
                "	ALPHA = 1.0; \n"
                "}");

            border_mat->set_shader(border_shader);
            border_mat->set_shader_parameter("albedo", Color(0.75, 0.75, 0.75, d.col.a / 3.0));

            rotate_gizmo[3] = Ref<ArrayMesh>(memnew(ArrayMesh()));
            rotate_gizmo[3]->add_surface_from_arrays(Mesh::PRIMITIVE_TRIANGLES, arrays);
            rotate_gizmo[3]->surface_set_material(0, border_mat);
        }
    }

    void
    Manipulator::init_scale_gizmo(const AxisData& d)
    {
        // Scale
        scale_gizmo[d.axis] = Ref<ArrayMesh>(memnew(ArrayMesh()));

        Ref<SurfaceTool> surftool = memnew(SurfaceTool());
        surftool->begin(Mesh::PRIMITIVE_TRIANGLES);

        // Cube arrow profile
        const int arrow_points = 6;
        Vector3 arrow[6] = {
            d.nivec * 0.0 + d.ivec * 0.0,
            d.nivec * 0.02 + d.ivec * 0.0,
            d.nivec * 0.02 + d.ivec * 1.0 * params.gizmo_scale_offset,
            d.nivec * 0.07 + d.ivec * 1.0 * params.gizmo_scale_offset,
            d.nivec * 0.07 + d.ivec * 1.11 * params.gizmo_scale_offset,
            d.nivec * 0.0 + d.ivec * 1.11 * params.gizmo_scale_offset,
        };

        int arrow_sides = 4;

        for (int k = 0; k < 4; k++)
        {
            Basis ma(d.ivec, Math_PI * 2 * float(k) / arrow_sides);
            Basis mb(d.ivec, Math_PI * 2 * float(k + 1) / arrow_sides);

            for (int j = 0; j < arrow_points - 1; j++)
            {
                Vector3 points[4] = {
                    ma.xform(arrow[j]),
                    mb.xform(arrow[j]),
                    mb.xform(arrow[j + 1]),
                    ma.xform(arrow[j + 1]),
                };
                surftool->add_vertex(points[0]);
                surftool->add_vertex(points[1]);
                surftool->add_vertex(points[2]);

                surftool->add_vertex(points[0]);
                surftool->add_vertex(points[2]);
                surftool->add_vertex(points[3]);
            }
        }

        surftool->set_material(d.mat);
        surftool->commit(scale_gizmo[d.axis]);
    }

    void
    Manipulator::init_scale_plane_gizmo(const AxisData& d)
    {
        // Plane Scale
        scale_plane_gizmo[d.axis] = Ref<ArrayMesh>(memnew(ArrayMesh()));

        Ref<SurfaceTool> surftool = memnew(SurfaceTool());
        surftool->begin(Mesh::PRIMITIVE_TRIANGLES);

        Vector3 vec = d.ivec2 - d.ivec3;
        Vector3 plane[4] = {vec * params.gizmo_plane_dst,
                            vec * params.gizmo_plane_dst + d.ivec2 * params.gizmo_plane_size,
                            vec * (params.gizmo_plane_dst + params.gizmo_plane_size),
                            vec * params.gizmo_plane_dst - d.ivec3 * params.gizmo_plane_size};

        Basis ma(d.ivec, Math_PI / 2);

        Vector3 points[4] = {
            ma.xform(plane[0]),
            ma.xform(plane[1]),
            ma.xform(plane[2]),
            ma.xform(plane[3]),
        };
        surftool->add_vertex(points[0]);
        surftool->add_vertex(points[1]);
        surftool->add_vertex(points[2]);

        surftool->add_vertex(points[0]);
        surftool->add_vertex(points[2]);
        surftool->add_vertex(points[3]);

        Ref<StandardMaterial3D> plane_mat = memnew(StandardMaterial3D());
        plane_mat->set_shading_mode(StandardMaterial3D::SHADING_MODE_UNSHADED);
        // plane_mat->set_on_top_of_alpha();
        plane_mat->set_transparency(StandardMaterial3D::TRANSPARENCY_ALPHA);
        plane_mat->set_cull_mode(StandardMaterial3D::CULL_DISABLED);
        plane_mat->set_albedo(d.col);
        plane_gizmo_color[d.axis] = plane_mat; // needed, so we can draw planes from both sides
        surftool->set_material(plane_mat);
        surftool->commit(scale_plane_gizmo[d.axis]);

        Ref<StandardMaterial3D> plane_mat_hl = plane_mat->duplicate();
        plane_mat_hl->set_albedo(d.col.from_hsv(d.col.get_h(), 0.25, 1.0, 1));
        plane_gizmo_color_hl[d.axis] =
            plane_mat_hl; // needed, so we can draw planes from both sides
    }

} // namespace godot
