#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "VideoRecorder.h"

#include <opencv2/imgcodecs.hpp>
#include <opencv2/opencv.hpp>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/DateTime.h>

#include <godot_cpp/classes/performance.hpp>

avi::recording::VideoRecorder::VideoRecorder(scene::CameraSet& cameras,
                                             core::SettingsProvider& settings,
                                             core::EventSubscriptionSystem& events) :
    cameras(cameras), settings(settings)
{
    events.subscribe("RECORD_VIDEO",
                     [this]()
                     {
                         if (not this->recording)
                         {
                             this->startRecording();
                         }
                         else
                         {
                             this->stopRecording();
                         }
                     });
}

void
avi::recording::VideoRecorder::run()
{
    DETOOLIFY;
    if (!this->recording)
        return;

    Recording& rec = *this->recording;

    std::chrono::steady_clock::time_point lastFrame = rec.lastFrameTime;
    std::chrono::steady_clock::time_point now = std::chrono::steady_clock::now();

    int64_t elapsed =
        std::chrono::duration_cast<std::chrono::milliseconds>(now - lastFrame).count();
    if (elapsed < MS_PER_FRAME)
        return;

    rec.lastFrameTime = now;

    godot::Ref<godot::Image> image = rec.camera->getCurrentScreenContent().image;
    image->convert(godot::Image::FORMAT_RGB8);

    if (image->get_width() != rec.frameSize.width || image->get_height() != rec.frameSize.height)
    {
        ARMARX_ERROR << "Screen size changed, aborting video recording" << std::endl;
        this->stopRecording();
        return;
    }

    godot::PackedByteArray data = image->get_data();

    cv::Mat in(rec.frameSize.height, rec.frameSize.width, CV_8UC3, (void*)data.ptr());
    cv::Mat out(rec.frameSize.height, rec.frameSize.width, CV_8UC3);
    cv::cvtColor(in, out, cv::COLOR_RGB2BGR);

    rec.writer << out;
}

void
avi::recording::VideoRecorder::startRecording()
{
    DETOOLIFY;
    auto* camera = this->cameras.getActiveCamera();

    std::filesystem::path directory = settings.getVideoDirectory();
    std::filesystem::create_directories(directory);

    armarx::DateTime now = armarx::DateTime::Now();
    std::string time = now.toDateTimeString();
    std::string filename = "avi_" + time + ".mp4";

    std::filesystem::path path = directory / filename;
    lastRecordingPath = path;

    godot::Vector2 size = camera->getScreenSize();
    int width = std::floor(size.x);
    int height = std::floor(size.y);

    this->recording = {.camera = this->cameras.getActiveCamera(),
                       .writer = cv::VideoWriter(path.string(),
                                                 cv::VideoWriter::fourcc('m', 'p', '4', 'v'),
                                                 FPS,
                                                 cv::Size(width, height)),
                       .frameSize = {width, height},
                       .lastFrameTime = std::chrono::steady_clock::now()};
}

void
avi::recording::VideoRecorder::stopRecording()
{
    DETOOLIFY;
    this->recording->writer.release();
    this->recording = std::nullopt;

    godot::OS* os = godot::OS::get_singleton();
    godot::PackedStringArray poolStringArray;
    poolStringArray.push_back("Video saved! Click 'Scene' > 'Open Video Directory'");
    poolStringArray.push_back(settings.getVideoDirectory().c_str());

    os->execute("notify-send", poolStringArray);
}

bool
avi::recording::VideoRecorder::isRecording()
{
    return recording.has_value();
}
std::filesystem::path
avi::recording::VideoRecorder::getLastRecordingPath()
{
    std::filesystem::path path = settings.getVideoDirectory();
    if (!lastRecordingPath.empty())
    {
        path = lastRecordingPath;
    }

    return path;
}
