#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "GraphicsController.h"

#include <utility>

#include <godot_cpp/classes/camera3d.hpp>
#include <godot_cpp/classes/environment.hpp>

avi::scene::GraphicsController::GraphicsController(godot::Node* root) : root(root)
{
    DETOOLIFY;
    world = root->get_node<godot::WorldEnvironment>("Control/3D/3D/World");
    viewport3D = root->get_node<godot::Viewport>("Control/3D/3D");
    light = root->get_node<godot::Light3D>("Control/3D/3D/Directional Light");
}

void
avi::scene::GraphicsController::activateProfile(avi::scene::Profile profile)
{
    DETOOLIFY;
    activeProfile = std::move(profile);
    update();
}

avi::scene::Profile&
avi::scene::GraphicsController::getActiveProfile()
{
    return activeProfile;
}

const std::string&
avi::scene::GraphicsController::getActiveProfileName()
{
    return activeProfile.getName();
}

static void
clamp(double* value, double min, double max)
{
    *value = std::clamp(*value, min, max);
}

void
avi::scene::GraphicsController::update()
{
    DETOOLIFY;

    godot::Viewport::MSAA msaa = activeProfile.msaa;
    if (msaa < 0 || msaa >= godot::Viewport::MSAA_MAX)
    {
        msaa = godot::Viewport::MSAA_DISABLED;
    }

    viewport3D->set_msaa_3d(msaa);
    root->get_viewport()->set_msaa_3d(msaa);

    float fov = godot::Math::clamp(static_cast<float>(activeProfile.fov), 1.0f, 179.0f);
    viewport3D->get_camera_3d()->set_fov(fov);
    activeProfile.fov = fov;

    auto environment = world->get_environment();

    environment->set_bg_color(activeProfile.backgroundColor);

    clamp(&activeProfile.ambientLight, 0.0, 100.0);
    environment->set_ambient_light_energy(static_cast<float>(activeProfile.ambientLight));

    clamp(&activeProfile.directionalLight, 0.0, 100.0);

    light->set_param(godot::Light3D::PARAM_ENERGY,
                     static_cast<float>(activeProfile.directionalLight));

    bool isLightEnabled = activeProfile.directionalLight > 0.0;
    light->set_visible(isLightEnabled);

    world->set_environment(environment);

    for (auto& [listener, callback] : changeListeners)
    {
        callback();
    }
}

avi::scene::DrawStyle
avi::scene::GraphicsController::getCurrentDrawStyle() const
{
    return activeProfile.drawStyle;
}

void
avi::scene::GraphicsController::registerChangeListener(void* listener,
                                                       const std::function<void()>& callback)
{
    DETOOLIFY;
    changeListeners.emplace(reinterpret_cast<uintptr_t>(listener), callback);
}

void
avi::scene::GraphicsController::unregisterChangeListener(void* listener)
{
    DETOOLIFY;
    auto it = changeListeners.find(reinterpret_cast<uintptr_t>(listener));
    if (it != changeListeners.end())
    {
        changeListeners.erase(it);
    }
}

bool
avi::scene::GraphicsController::isHeadlightEnabled() const
{
    return activeProfile.cameraHeadlight;
}

bool
avi::scene::GraphicsController::isCoordinateSystemEnabled() const
{
    return activeProfile.showCoordinateLines;
}

godot::Color
avi::scene::GraphicsController::getBackgroundColor() const
{
    return activeProfile.backgroundColor;
}

double
avi::scene::GraphicsController::getHeadlightStrength() const
{
    return activeProfile.headlightStrength;
}
