#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "LineElementNode.h"

#include <godot_cpp/classes/cylinder_mesh.hpp>
#include <godot_cpp/classes/cylinder_shape3d.hpp>
#include <godot_cpp/classes/sphere_mesh.hpp>
#include <godot_cpp/classes/sphere_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::LineElementNode::_bind_methods()
{
    DETOOLIFY;

}

void
avi::scene::LineElementNode::_ready()
{
    DETOOLIFY;
}

void
avi::scene::LineElementNode::setStartAttributes(std::string name, avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(avi::connection::SentElementType::LINE);
    setName(name);
    setLayer(layer);

    staticBody = memnew(godot::StaticBody3D);
    this->add_child(staticBody);

    start = memnew(godot::MeshInstance3D);
    staticBody->add_child(start);
    end = memnew(godot::MeshInstance3D);
    staticBody->add_child(end);

    godot::Mesh* meshSphere = memnew(godot::SphereMesh);
    start->set_mesh(meshSphere);
    end->set_mesh(meshSphere);

    line = memnew(godot::MeshInstance3D);
    staticBody->add_child(line);
    godot::Mesh* meshLine = memnew(godot::CylinderMesh);
    line->set_mesh(meshLine);

    material = memnew(godot::StandardMaterial3D);
    line->get_mesh()->surface_set_material(0, material);
    start->get_mesh()->surface_set_material(0, material);
    end->get_mesh()->surface_set_material(0, material);

    collisionLine = memnew(godot::CollisionShape3D);
    collisionLine->set_shape(memnew(godot::CylinderShape3D));
    staticBody->add_child(collisionLine);
}

void
avi::scene::LineElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quaternion rotation,
                                              godot::Color color)
{
    DETOOLIFY;
    setInteractionFlag(flag);
    setInteractions(std::move(interactions));

    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
}

void
avi::scene::LineElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto lineHandle = IceInternal::Handle<armarx::viz::data::ElementLine>::dynamicCast(elementInfo);

    if (!godot::Math::is_equal_approx(radius, lineHandle->lineWidth))
    {
        updateRadius(lineHandle->lineWidth);
    }

    godot::Vector3 from = {lineHandle->from.e0, lineHandle->from.e1, lineHandle->from.e2};
    godot::Vector3 to = {lineHandle->to.e0, lineHandle->to.e1, lineHandle->to.e2};

    if (!avi::core::utilities::is_equal_approx(a, from) ||
        !avi::core::utilities::is_equal_approx(b, to))
    {
        updateLine(from, to);
    }

    updateAttributes(lineHandle->interaction.enableFlags,
                     lineHandle->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(lineHandle->pose),
                     avi::scene::Converter::getScale(lineHandle->scale),
                     avi::scene::Converter::getRotation(lineHandle->pose),
                     avi::scene::Converter::getColor(lineHandle->color));
}

avi::scene::LineElementNode*
avi::scene::LineElementNode::newLineElementNode(std::string name, avi::core::Layer* layer)
{
    LineElementNode* node = memnew(LineElementNode);
    node->setStartAttributes(name, layer);
    return node;
}

godot::AABB
avi::scene::LineElementNode::getVolume()
{
    godot::AABB aabb(a / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR, godot::Vector3(0, 0, 0));
    aabb.expand_to(b / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    aabb.position += staticBody->get_transform().origin;
    aabb.size *= staticBody->get_scale() * 0.5 * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;

    return aabb;
}

godot::Transform3D
avi::scene::LineElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(0, 0, 0) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

std::vector<godot::Node3D*>
avi::scene::LineElementNode::getSpatial()
{
    std::vector<godot::Node3D*> spatial;

    spatial.push_back(start);
    spatial.push_back(end);
    spatial.push_back(line);

    return spatial;
}

void
avi::scene::LineElementNode::updateRadius(float newRadius)
{
    DETOOLIFY;
    radius = newRadius;

    godot::Vector3 scale = {newRadius, line->get_scale().y, newRadius};
    line->set_scale(scale);
    collisionLine->set_scale(scale);

    float sphereRadius = radius * 1.5f;
    godot::Vector3 sphereScale = {sphereRadius, sphereRadius, sphereRadius};

    start->set_scale(sphereScale);
    end->set_scale(sphereScale);
}

void
avi::scene::LineElementNode::updateLine(godot::Vector3 from, godot::Vector3 to)
{
    DETOOLIFY;
    a = from;
    b = to;

    start->set_position(from);
    end->set_position(to);

    godot::Vector3 direction = to - from;
    godot::Vector3 scale = line->get_scale();
    godot::Vector3 mid = from + direction / 2;

    godot::Quaternion rotation(godot::Vector3(0, 1, 0), direction.normalized());
    godot::Transform3D transform;
    transform.basis *= rotation;
    line->set_transform(transform);

    line->set_position(mid);
    line->set_scale({scale.x, direction.length() / 2, scale.z});

    collisionLine->set_transform(line->get_transform());
}

void
avi::scene::LineElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    staticBody->set_position(position);
}

void
avi::scene::LineElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    staticBody->set_rotation(rotation.get_euler());
}

void
avi::scene::LineElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::LineElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    material->set_albedo(color);

    if (color.a < 1)
    {
        material->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
    }
}