#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "MeshElementNode.h"

#include <iostream>

#include <RobotAPI/interface/ArViz/Elements.h>

#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/array_mesh.hpp>
#include <godot_cpp/classes/shape3d.hpp>
#include <godot_cpp/classes/concave_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>

avi::scene::MeshElementNode::MeshElementNode() = default;
avi::scene::MeshElementNode::~MeshElementNode() = default;

godot::AABB
avi::scene::MeshElementNode::getVolume()
{
    godot::AABB correctedAABB = meshInstance->get_aabb();
    correctedAABB.position = staticBody->get_global_transform().origin;
    correctedAABB.size *= meshInstance->get_scale() * staticBody->get_scale() * 0.5;

    return correctedAABB;
}

void
avi::scene::MeshElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);

    staticBody->set_transform(transform);
}

void
avi::scene::MeshElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::MeshElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::MeshElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    auto* mat = cast_to<godot::StandardMaterial3D>(meshInstance->get_mesh()->surface_get_material(0).ptr());
    mat->set_albedo(color);
    if (color.a < 1)
    {
        mat->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
    }
}

godot::Vector3
avi::scene::MeshElementNode::getPosition()
{
    return staticBody->get_position();
}

godot::Vector3
avi::scene::MeshElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::MeshElementNode::getScale()
{
    return staticBody->get_scale();
}

void
avi::scene::MeshElementNode::setStartAttributes(godot::Mesh* mesh,
                                                avi::connection::SentElementType type,
                                                std::string name,
                                                avi::core::Layer* layer)
{
    DETOOLIFY;
    staticBody = memnew(godot::StaticBody3D);
    collisionShape = memnew(godot::CollisionShape3D);
    meshInstance = memnew(godot::MeshInstance3D);

    add_child(staticBody);
    staticBody->add_child(collisionShape);

    staticBody->add_child(meshInstance);

    godot::StandardMaterial3D* material = memnew(godot::StandardMaterial3D);
    material->set_shading_mode(godot::StandardMaterial3D::SHADING_MODE_PER_PIXEL);

    meshInstance->set_mesh(godot::Ref<godot::Mesh>{mesh});
    meshInstance->get_mesh()->surface_set_material(0, material);

    setMesh(mesh);
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::MeshElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quaternion rotation,
                                              godot::Color color)
{
    DETOOLIFY;
    setInteractionFlag(flag);
    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
}

godot::Mesh*
avi::scene::MeshElementNode::getMesh()
{
    return meshInstance->get_mesh().ptr();
}

void
avi::scene::MeshElementNode::setMesh(godot::Mesh* mesh)
{
    DETOOLIFY;
    meshInstance->set_mesh(godot::Ref<godot::Mesh>{mesh});
}

void
avi::scene::MeshElementNode::useVertexColor(bool isUsed)
{
    DETOOLIFY;
    auto* mat = cast_to<godot::StandardMaterial3D>(meshInstance->get_mesh()->surface_get_material(0).ptr());
    mat->set_flag(godot::StandardMaterial3D::FLAG_ALBEDO_FROM_VERTEX_COLOR, isUsed);
    mat->set_transparency(isUsed ? godot::StandardMaterial3D::TRANSPARENCY_ALPHA : godot::StandardMaterial3D::TRANSPARENCY_DISABLED);
    mat->set_diffuse_mode(godot::StandardMaterial3D::DIFFUSE_LAMBERT_WRAP);
}

void
avi::scene::MeshElementNode::setTwoSided()
{
    DETOOLIFY;
    auto* mat = cast_to<godot::StandardMaterial3D>(meshInstance->get_mesh()->surface_get_material(0).ptr());
    mat->set_cull_mode(godot::StandardMaterial3D::CULL_DISABLED);
}

void
avi::scene::MeshElementNode::setShape(godot::Mesh* mesh)
{
    DETOOLIFY;
    godot::Ref<godot::Shape3D> shape = mesh->create_trimesh_shape();
    this->collisionShape->set_shape(shape);
    collisionShape->set_scale(meshInstance->get_scale());
}

void
avi::scene::MeshElementNode::setShape(godot::Shape3D* shape)
{
    DETOOLIFY;
    this->collisionShape->set_shape(shape);
}

godot::Transform3D
avi::scene::MeshElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

std::vector<godot::Node3D*>
avi::scene::MeshElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    vec.push_back(meshInstance);
    return vec;
}

godot::MeshInstance3D*
avi::scene::MeshElementNode::getMeshInstance()
{
    return meshInstance;
}

godot::CollisionShape3D*
avi::scene::MeshElementNode::getCollisionShape()
{
    return collisionShape;
}
