#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "MultiMeshElementNode.h"

#include <iostream>

#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/box_mesh.hpp>
#include <godot_cpp/classes/convex_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>


avi::scene::MultiMeshElementNode::MultiMeshElementNode() = default;
avi::scene::MultiMeshElementNode::~MultiMeshElementNode() = default;

godot::AABB
avi::scene::MultiMeshElementNode::getVolume()
{
    godot::AABB AABB = meshInstances.front()->get_aabb();
    for (size_t i = 1; i < meshInstances.size(); ++i)
    {
        AABB.merge_with(meshInstances[i]->get_aabb());
    }
    AABB.position = meshInstances.front()->get_global_transform().origin * 0.001;
    AABB.size = AABB.size * staticBody->get_scale();

    return AABB;
}

void
avi::scene::MultiMeshElementNode::setPosition(godot::Vector3 position, size_t index)
{
    DETOOLIFY;

    auto transform = meshInstances.at(index)->get_transform();
    transform.set_origin(position * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);
    meshInstances.at(index)->set_transform(transform);

    if (index < collisionShapes.size())
    {
        collisionShapes.at(index)->set_transform(transform);
    }
}

void
avi::scene::MultiMeshElementNode::setScale(godot::Vector3 scale, size_t index)
{
    DETOOLIFY;
    staticBody->set_scale({0.001, 0.001, 0.001});
    meshInstances.at(index)->set_scale(scale * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);
    if (index < collisionShapes.size())
    {
        collisionShapes.at(index)->set_scale(scale * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);
    }
}

void
avi::scene::MultiMeshElementNode::setRotation(godot::Quaternion rotation, size_t index)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = meshInstances.at(index)->get_transform().origin;

    meshInstances.at(index)->set_transform(quaternionTransform);
    if (index < collisionShapes.size())
    {
        quaternionTransform.origin = collisionShapes.at(index)->get_transform().origin;
        collisionShapes.at(index)->set_transform(quaternionTransform);
    }
}

void
avi::scene::MultiMeshElementNode::setColor(godot::Color color, size_t index)
{
    DETOOLIFY;
    godot::StandardMaterial3D* mat =
        cast_to<godot::StandardMaterial3D>(meshInstances.at(index)->get_mesh()->surface_get_material(0).ptr());
    mat->set_albedo(color);
    if (color.a < 1)
    {
        mat->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
    }
}

std::vector<godot::Vector3>
avi::scene::MultiMeshElementNode::getPosition()
{
    std::vector<godot::Vector3> result;
    for (auto instance : meshInstances)
    {
        result.push_back(instance->get_position());
    }
    return result;
}

std::vector<godot::Vector3>
avi::scene::MultiMeshElementNode::getRotation()
{
    std::vector<godot::Vector3> result;
    for (auto instance : meshInstances)
    {
        result.push_back(instance->get_rotation());
    }
    return result;
}

std::vector<godot::Vector3>
avi::scene::MultiMeshElementNode::getScale()
{
    std::vector<godot::Vector3> result;
    for (auto instance : meshInstances)
    {
        result.push_back(instance->get_scale());
    }
    return result;
}

void
avi::scene::MultiMeshElementNode::setStartAttributes(std::vector<godot::Mesh*> meshes,
                                                     avi::connection::SentElementType type,
                                                     std::string name,
                                                     avi::core::Layer* layer)
{
    DETOOLIFY;
    staticBody = memnew(godot::StaticBody3D);
    add_child(staticBody);
    setType(type);
    setName(name);
    setLayer(layer);
    setMesh(meshes);
}

void
avi::scene::MultiMeshElementNode::updateAttributes(int flag,
                                                   std::vector<std::string> interactions,
                                                   godot::Vector3 position,
                                                   godot::Vector3 scale,
                                                   godot::Quaternion rotation,
                                                   godot::Color color)
{
    DETOOLIFY;
    setInteractionFlag(flag);
    if (!shapeChecked)
    {
        if (ElementNode::isSelectable())
        {
            setShape();
        }
        shapeChecked = true;
    }
    setInteractions(interactions);
    for (size_t i = 0; i < meshInstances.size(); i++)
    {
        setPosition(position, i);
        setScale(scale, i);
        setRotation(rotation, i);
        setColor(color, i);
    }
}

godot::Mesh*
avi::scene::MultiMeshElementNode::getMesh(int index)
{
    return meshInstances.at(index)->get_mesh().ptr();
}

void
avi::scene::MultiMeshElementNode::setMesh(std::vector<godot::Mesh*> meshes)
{
    DETOOLIFY;
    for (auto mesh : meshes)
    {
        godot::MeshInstance3D* mesh_instance = memnew(godot::MeshInstance3D);
        mesh_instance->set_mesh(mesh);
        meshInstances.push_back(mesh_instance);
        godot::StandardMaterial3D* material = memnew(godot::StandardMaterial3D);
        material->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
        material->set_shading_mode(godot::StandardMaterial3D::SHADING_MODE_PER_PIXEL);
        mesh_instance->get_mesh()->surface_set_material(0, material);
        staticBody->add_child(mesh_instance);
    }
}

void
avi::scene::MultiMeshElementNode::setTwoSided(int index)
{
    DETOOLIFY;
    auto* mat =
        cast_to<godot::StandardMaterial3D>(meshInstances.at(index)->get_mesh()->surface_get_material(0).ptr());
    mat->set_cull_mode(godot::StandardMaterial3D::CULL_DISABLED);
}

void
avi::scene::MultiMeshElementNode::setVisibility(int index, bool visible)
{
    DETOOLIFY;
    meshInstances.at(index)->set_visible(visible);
}

godot::Transform3D
avi::scene::MultiMeshElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

void
avi::scene::MultiMeshElementNode::setShape()
{
    DETOOLIFY;
    for (auto mesh : this->meshInstances)
    {
        godot::CollisionShape3D* collisionShape = memnew(godot::CollisionShape3D);
        godot::Ref<godot::Shape3D> shape = mesh->get_mesh()->create_convex_shape();

        collisionShape->set_shape(shape);

        collisionShapes.push_back(collisionShape);
        staticBody->add_child(collisionShape);
    }
}

std::vector<godot::Node3D*>
avi::scene::MultiMeshElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    for (auto* meshInstance : meshInstances)
    {
        vec.push_back(meshInstance);
    }
    return vec;
}

void
avi::scene::MultiMeshElementNode::update(
    IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    ElementNode::update(elementInfo);
}
