#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "ObjectElementNode.h"

#include <iostream>
#include <utility>

#include <VirtualRobot/ManipulationObject.h>

#include <godot_cpp/classes/box_mesh.hpp>
#include <godot_cpp/classes/standard_material3d.hpp>
#include <godot_cpp/classes/concave_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::ObjectElementNode::_bind_methods()
{
    DETOOLIFY;

}

godot::AABB
avi::scene::ObjectElementNode::getVolume()
{
    godot::Ref<godot::ArrayMesh> mesh = meshInstance->get_mesh();

    if (mesh.is_valid() && mesh->get_surface_count() == 0)
    {
        godot::AABB emptyAABB;

        emptyAABB.position = staticBody->get_global_transform().origin;
        emptyAABB.size = godot::Vector3(1, 1, 1);

        return emptyAABB;
    }

    godot::AABB AABB = meshInstance->get_aabb();

    AABB.position = staticBody->get_global_transform().origin;
    AABB.size = AABB.size * staticBody->get_scale() * 0.5;

    return AABB;
}

avi::scene::ObjectElementNode::ObjectElementNode()
{
    DETOOLIFY;
}

void
avi::scene::ObjectElementNode::_ready()
{
    DETOOLIFY;
}

void
avi::scene::ObjectElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::ObjectElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::ObjectElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

godot::Vector3
avi::scene::ObjectElementNode::getPosition()
{
    return staticBody->get_position();
}

godot::Vector3
avi::scene::ObjectElementNode::getRotation()
{
    return staticBody->get_rotation();
}

godot::Vector3
avi::scene::ObjectElementNode::getScale()
{
    return staticBody->get_scale();
}

avi::scene::ObjectElementNode*
avi::scene::ObjectElementNode::newObjectElementNode(avi::core::ModelLoader* loader,
                                                    avi::connection::SentElementType type,
                                                    std::string name,
                                                    avi::core::Layer* layer)
{

    ObjectElementNode* node = memnew(ObjectElementNode);
    node->staticBody = memnew(godot::StaticBody3D);
    node->add_child(node->staticBody);
    node->loader = loader;

    node->setStartAttributes(type, std::move(name), layer);

    return node;
}

void
avi::scene::ObjectElementNode::setStartAttributes(avi::connection::SentElementType type,
                                                  std::string name,
                                                  avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(type);
    setName(name);
    setLayer(layer);
}

void
avi::scene::ObjectElementNode::updateAttributes(std::string fileName,
                                                int flag,
                                                std::vector<std::string> interactions,
                                                godot::Vector3 position,
                                                godot::Vector3 scale,
                                                godot::Quaternion rotation,
                                                godot::Color color,
                                                ModelDrawStyle drawStyle)
{
    setInteractionFlag(flag);
    if (this->fileName != fileName)
    {
        if (meshInstance != nullptr && collisionShape != nullptr)
        {
            staticBody->remove_child(meshInstance);
            staticBody->remove_child(collisionShape);
        }

        this->fileName = fileName;
        std::string fullName;
        godot::Ref<godot::Mesh> object = loader->loadModel(fileName, fullName);

        for (int64_t surface = 0; surface < object->get_surface_count(); surface++)
        {
            godot::Ref<godot::StandardMaterial3D> material = object->surface_get_material(surface);

            if (material.is_valid())
            {
                if (material->get_albedo().a != 1)
                {
                    material->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
                }
            }
        }

        {
            meshInstance = memnew(godot::MeshInstance3D);
            meshInstance->set_mesh(object);
            staticBody->add_child(meshInstance);
        }
        {
            collisionShape = memnew(godot::CollisionShape3D);
            staticBody->add_child(collisionShape);
        }

        if (!shapeChecked)
        {
            setShape(loader->getObject(fullName));
            shapeChecked = true;
        }
    }

    if (currentDrawStyle != drawStyle || currentColor != color)
    {
        currentDrawStyle = drawStyle;
        currentColor = color;

        setOverrideColor(color, drawStyle);
    }

    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale);
}

void
avi::scene::ObjectElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto object = IceInternal::Handle<armarx::viz::data::ElementObject>::dynamicCast(elementInfo);
    updateAttributes(object->filename,
                     object->interaction.enableFlags,
                     object->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(object->pose),
                     godot::Vector3(object->scale.e0, object->scale.e1, object->scale.e2),
                     avi::scene::Converter::getRotation(object->pose),
                     avi::scene::Converter::getColor(object->color),
                     ModelDrawStyle(object->drawStyle));
}

godot::Transform3D
avi::scene::ObjectElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);
    return transform;
}

void
avi::scene::ObjectElementNode::setShape(VirtualRobot::ManipulationObjectPtr visualization)
{
    DETOOLIFY;
    assert(loader);

    if (visualization && visualization->getCollisionModel() &&
        visualization->getCollisionModel()->getTriMeshModel())
    {
        std::optional<godot::Ref<godot::Shape3D>> shape =
            loader->getCollisionShapeCache().get(visualization);

        if (shape)
        {
            collisionShape->set_shape(shape.value());
        }
        else
        {
            godot::Ref<godot::Mesh> obj = avi::core::ModelLoader::createObjectFromTriMesh(
                visualization->getCollisionModel()->getTriMeshModel());
            godot::Ref<godot::Shape3D> newShape = obj->create_trimesh_shape();
            collisionShape->set_shape(newShape);

            loader->getCollisionShapeCache().emplace(visualization, newShape);
        }
    }
    else
    {
        std::optional<godot::Ref<godot::Shape3D>> shape =
            loader->getShapeCache().get(meshInstance->get_mesh().ptr());

        if (shape)
        {
            collisionShape->set_shape(shape.value());
        }
        else
        {
            godot::Ref<godot::Shape3D> newShape = meshInstance->get_mesh()->create_trimesh_shape();
            loader->getShapeCache().emplace(meshInstance->get_mesh().ptr(), {newShape.ptr()});

            collisionShape->set_shape(newShape);
        }
    }
}

std::vector<godot::Node3D*>
avi::scene::ObjectElementNode::getSpatial()
{
    return {meshInstance};
}

void
avi::scene::ObjectElementNode::setOverrideColor(godot::Color color, ModelDrawStyle drawStyle)
{
    DETOOLIFY;
    std::string fullName;
    godot::Ref<godot::Mesh> object = loader->loadModel(fileName, fullName);

    if (drawStyle.isColorOverriden())
    {
        for (int64_t surface = 0; surface < object->get_surface_count(); surface++)
        {
            godot::Ref<godot::StandardMaterial3D> overrideMaterial =
                meshInstance->get_material_override().is_null()
                    ? memnew(godot::StandardMaterial3D)
                    : meshInstance->get_material_override();

            overrideMaterial->set_albedo(color);
            overrideMaterial->set_flag(godot::StandardMaterial3D::FLAG_ALBEDO_FROM_VERTEX_COLOR, true);

            overrideMaterial->set_blend_mode(godot::StandardMaterial3D::BLEND_MODE_MIX);

            overrideMaterial->set_transparency(color.a != 1 ? godot::StandardMaterial3D::TRANSPARENCY_ALPHA : godot::StandardMaterial3D::TRANSPARENCY_DISABLED);

            godot::Ref<godot::StandardMaterial3D> material = object->surface_get_material(surface);

            if (material.is_valid())
            {
                overrideMaterial->set_texture(
                    godot::StandardMaterial3D::TEXTURE_ALBEDO,
                    material->get_texture(godot::StandardMaterial3D::TEXTURE_ALBEDO));
            }

            meshInstance->set_material_override(overrideMaterial);
        }
    }
    else
    {
        meshInstance->set_material_override(nullptr);
    }
}