#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "PathElementNode.h"

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/scene/Converter.h>
#include <godot_cpp/classes/collision_shape3d.hpp>
#include <godot_cpp/classes/cylinder_mesh.hpp>
#include <godot_cpp/classes/cylinder_shape3d.hpp>
#include <godot_cpp/classes/mesh.hpp>
#include <godot_cpp/classes/multi_mesh.hpp>
#include <godot_cpp/classes/multi_mesh_instance3d.hpp>
#include <godot_cpp/classes/sphere_mesh.hpp>
#include <godot_cpp/classes/static_body3d.hpp>

void
avi::scene::PathElementNode::_bind_methods()
{
    DETOOLIFY;

}

void
avi::scene::PathElementNode::_ready()
{
    DETOOLIFY;
}

void
avi::scene::PathElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto path = IceInternal::Handle<armarx::viz::data::ElementPath>::dynamicCast(elementInfo);

    bool allPointsApproxEqual = points.size() == path->points.size();

    std::vector<godot::Vector3> newPoints;
    newPoints.reserve(path->points.size());
    for (size_t i = 0; i < path->points.size(); ++i)
    {
        godot::Vector3 point = {path->points[i].e0, path->points[i].e1, path->points[i].e2};
        newPoints.push_back(point);

        if (allPointsApproxEqual && i < points.size())
        {
            allPointsApproxEqual &= avi::core::utilities::is_equal_approx(points[i], point);
        }
    }

    if (!godot::Math::is_equal_approx(radius, path->lineWidth) || !allPointsApproxEqual)
    {
        radius = path->lineWidth;
        points = std::move(newPoints);

        buildPath();
    }

    updateAttributes(path->interaction.enableFlags,
                     path->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(path->pose),
                     avi::scene::Converter::getScale(path->scale),
                     avi::scene::Converter::getRotation(path->pose),
                     avi::scene::Converter::getColor(path->color));
}

avi::scene::PathElementNode*
avi::scene::PathElementNode::newPathElementNode(std::string name, avi::core::Layer* layer)
{
    PathElementNode* node = memnew(PathElementNode);
    node->setStartAttributes(name, layer);
    return node;
}

void
avi::scene::PathElementNode::setStartAttributes(std::string name, avi::core::Layer* layer)
{
    DETOOLIFY;
    setType(avi::connection::SentElementType::PATH);
    setName(name);
    setLayer(layer);

    staticBody = memnew(godot::StaticBody3D);
    this->add_child(staticBody);

    godot::MultiMeshInstance3D* lines = memnew(godot::MultiMeshInstance3D);
    staticBody->add_child(lines);
    spatials.push_back(lines);

    meshLine = memnew(godot::MultiMesh);
    meshLine->set_mesh(memnew(godot::CylinderMesh));
    lines->set_multimesh(meshLine);

    meshLine->set_transform_format(godot::MultiMesh::TRANSFORM_3D);
    meshLine->set_use_colors(false);
    meshLine->set_use_custom_data(false);

    godot::MultiMeshInstance3D* spheres = memnew(godot::MultiMeshInstance3D);
    staticBody->add_child(spheres);
    spatials.push_back(spheres);

    meshSphere = memnew(godot::MultiMesh);
    meshSphere->set_mesh(memnew(godot::SphereMesh));
    spheres->set_multimesh(meshSphere);

    meshSphere->set_transform_format(godot::MultiMesh::TRANSFORM_3D);
    meshLine->set_use_colors(false);
    meshLine->set_use_custom_data(false);

    material = memnew(godot::StandardMaterial3D);
    lines->set_material_override(material);
    spheres->set_material_override(material);
}

void
avi::scene::PathElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quaternion rotation,
                                              godot::Color color)
{
    DETOOLIFY;
    setInteractionFlag(flag);
    setInteractions(std::move(interactions));

    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
}

godot::AABB
avi::scene::PathElementNode::getVolume()
{
    bool noBaseAABB = true;
    godot::AABB AABB;

    godot::Vector3 pathPosition = staticBody->get_global_transform().origin;
    godot::Vector3 pathScale = staticBody->get_scale();
    godot::Basis changeOfBasisMatrix = this->get_global_transform().basis;

    for (auto point : points)
    {
        if (noBaseAABB)
        {
            AABB.position = changeOfBasisMatrix.xform(point) * pathScale + pathPosition;
            noBaseAABB = false;
        }
        else
        {
            godot::AABB tempAABB;
            tempAABB.position = changeOfBasisMatrix.xform(point) * pathScale + pathPosition;
            AABB.merge_with(tempAABB);
        }
    }

    return AABB;
}

godot::Transform3D
avi::scene::PathElementNode::getTransform()
{
    godot::Transform3D transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3(1, 1, 1) * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

std::vector<godot::Node3D*>
avi::scene::PathElementNode::getSpatial()
{
    return spatials;
}

void
avi::scene::PathElementNode::buildPath()
{
    DETOOLIFY;
    buildLines();
    buildSpheres();
}

void
avi::scene::PathElementNode::buildLines()
{
    DETOOLIFY;
    const size_t perInstanceData = 12;
    size_t lineCount = points.size() - 1;
    lineCount = !points.empty() ? lineCount : 0;

    int requiredSize = static_cast<int>(lineCount * perInstanceData);
    if (lineData.size() != requiredSize)
    {
        lineData.resize(requiredSize);
    }

    godot::Vector3 p0 = !points.empty() ? points.front() : godot::Vector3(0, 0, 0);
    minPoint = p0;
    maxPoint = p0;

    real_t* data = lineData.ptrw();

    std::vector<godot::CollisionShape3D*> newShapes;
    newShapes.reserve(lineCount);
    godot::Shape3D* shape = memnew(godot::CylinderShape3D);

    for (ulong lineIndex = 0; lineIndex < lineCount; lineIndex++)
    {
        godot::Vector3& a = points[lineIndex];
        godot::Vector3& b = points[lineIndex + 1];
        real_t* currentData = data + (lineIndex * perInstanceData);

        size_t i = 0;

        godot::Vector3 direction = b - a;
        godot::Vector3 mid = a + direction / 2;

        godot::Transform3D rotate;
        rotate.basis *= godot::Quaternion(godot::Vector3(0, 1, 0), direction.normalized());

        godot::Transform3D scale;
        scale.scale({radius, direction.length() / 2, radius});

        godot::Basis basis = (rotate * scale).basis;

        currentData[i++] = basis.rows[0][0];
        currentData[i++] = basis.rows[0][1];
        currentData[i++] = basis.rows[0][2];
        currentData[i++] = mid.x;

        currentData[i++] = basis.rows[1][0];
        currentData[i++] = basis.rows[1][1];
        currentData[i++] = basis.rows[1][2];
        currentData[i++] = mid.y;

        currentData[i++] = basis.rows[2][0];
        currentData[i++] = basis.rows[2][1];
        currentData[i++] = basis.rows[2][2];
        currentData[i++] = mid.z;

        auto considerPointBounds = [&](const godot::Vector3& point)
        {
            minPoint = {std::min(minPoint.x, point.x),
                        std::min(minPoint.y, point.y),
                        std::min(minPoint.z, point.z)};

            maxPoint = {std::max(maxPoint.x, point.x),
                        std::max(maxPoint.y, point.y),
                        std::max(maxPoint.z, point.z)};
        };

        considerPointBounds(a);
        considerPointBounds(b);

        godot::CollisionShape3D* current;

        if (lineIndex < shapes.size())
        {
            current = shapes[lineIndex];
        }
        else
        {
            current = memnew(godot::CollisionShape3D);
            staticBody->add_child(current);
            current->set_shape(shape);
        }

        current->set_transform(rotate * scale);
        current->set_position(mid);

        newShapes.push_back(current);
    }

    meshLine->set_instance_count(static_cast<int64_t>(lineCount));
    meshLine->set_buffer(lineData);

    for (size_t index = newShapes.size(); index < shapes.size(); index++)
    {
        shapes[index]->queue_free();
    }

    shapes = std::move(newShapes);
}

void
avi::scene::PathElementNode::buildSpheres()
{
    DETOOLIFY;
    const size_t perInstanceData = 12;
    size_t sphereCount = points.size();

    int requiredSize = static_cast<int>(sphereCount * perInstanceData);
    if (sphereData.size() != requiredSize)
    {
        sphereData.resize(requiredSize);
    }

    real_t* data = sphereData.ptrw();

    for (ulong sphereIndex = 0; sphereIndex < sphereCount; sphereIndex++)
    {
        godot::Vector3& point = points[sphereIndex];
        real_t* currentData = data + (sphereIndex * perInstanceData);

        size_t i = 0;

        godot::Basis basis;
        basis.scale(godot::Vector3(radius, radius, radius) * 1.5f);

        currentData[i++] = basis.rows[0][0];
        currentData[i++] = basis.rows[0][1];
        currentData[i++] = basis.rows[0][2];
        currentData[i++] = point.x;

        currentData[i++] = basis.rows[1][0];
        currentData[i++] = basis.rows[1][1];
        currentData[i++] = basis.rows[1][2];
        currentData[i++] = point.y;

        currentData[i++] = basis.rows[2][0];
        currentData[i++] = basis.rows[2][1];
        currentData[i++] = basis.rows[2][2];
        currentData[i++] = point.z;
    }

    meshSphere->set_instance_count(static_cast<int32_t>(sphereCount));
    meshSphere->set_buffer(sphereData);
}

void
avi::scene::PathElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    staticBody->set_position(position);
}

void
avi::scene::PathElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::PathElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::PathElementNode::setColor(godot::Color color)
{
    DETOOLIFY;
    material->set_albedo(color);

    if (color.a < 1)
    {
        material->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
    }
}
