#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "PlayfulControlScheme.h"

#include <utility>

#include <godot_cpp/godot.hpp>
#include <godot_cpp/classes/input.hpp>
#include <godot_cpp/classes/input_event.hpp>
#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/input_event_mouse_button.hpp>
#include <godot_cpp/classes/input_event_mouse_motion.hpp>
#include <godot_cpp/classes/scene_tree.hpp>
#include <godot_cpp/classes/control.hpp>

#include <arviz_godot/lib/avi/core/Utilities.h>

void
avi::scene::PlayfulControlScheme::_bind_methods()
{
    DETOOLIFY;
}

avi::scene::PlayfulControlScheme::PlayfulControlScheme()
{
    viewCamera = nullptr;
    speedModifier = 0.0f;
    forwardMovement = 0.0f;

    inputState = std::vector<bool>(MAX_INPUT_INDEX, false);

    inputKeys.push_back("avi_camera_forward");
    inputKeys.push_back("avi_camera_backward");
    inputKeys.push_back("avi_camera_left");
    inputKeys.push_back("avi_camera_right");
    inputKeys.push_back("avi_camera_up");
    inputKeys.push_back("avi_camera_down");
    inputKeys.push_back("avi_camera_roll_left");
    inputKeys.push_back("avi_camera_roll_right");
    inputKeys.push_back("avi_camera_speed_down");
    inputKeys.push_back("avi_camera_speed_up");
}

void
avi::scene::PlayfulControlScheme::setup(avi::core::SettingsProvider& settings,
                                        avi::scene::VisualizedScene& scene,
                                        bool gravity)
{
    this->settings = &settings;
    this->scene = &scene;
    this->gravity = gravity;
}

void
avi::scene::PlayfulControlScheme::_ready()
{
    DETOOLIFY;
    attachment = get_node<godot::Node3D>("Attachment");
    head = get_node<godot::Node3D>("Attachment/Head");
}

void
avi::scene::PlayfulControlScheme::_unhandled_input(const godot::Ref<godot::InputEvent>& event)
{
    DETOOLIFY;
    if (!isActive()) return;

    auto* mouseMotionEvent = cast_to<godot::InputEventMouseMotion>(event.ptr());
    if (mouseMotionEvent)
    {
        mouseDeltaAccumulator += mouseMotionEvent->get_relative();
    }

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent)
    {
        auto idx = mouseButtonEvent->get_button_index();

        if (idx == godot::MOUSE_BUTTON_WHEEL_UP)
        {
            forwardMovement += 1.0f;
        }

        if (idx == godot::MOUSE_BUTTON_WHEEL_DOWN)
        {
            forwardMovement -= 1.0f;
        }
    }

    for (auto i = 0; i < MAX_INPUT_INDEX; i++)
    {
        if (event->is_action_pressed(inputKeys[i]))
            inputState[i] = true;
    }
}

template<typename V>
static V lerp(V a, V b, double factor)
{
    double t = std::clamp( factor / 10000.0, 0.0, 1.0);
    return a + (b - a) * static_cast<float>(t);
}

void
avi::scene::PlayfulControlScheme::_process(double delta)
{
    DETOOLIFY;
    if (!isActive()) return;

    if (isTraveling)
    {
        doTravel(delta);
        return;
    }

    godot::Input* input = godot::Input::get_singleton();

    bool rotate = input->is_action_pressed("avi_camera_mouse") || viewCamera->isMouseLocked();
    mouseDeltaAccumulator = rotate ? mouseDeltaAccumulator : godot::Vector2();
    smoothMouseDelta = lerp(smoothMouseDelta, mouseDeltaAccumulator, settings->getSmoothingFactorPlayful());
    godot::Vector2 mouseDelta = settings->isPlayfulInputSmoothed() ? smoothMouseDelta : mouseDeltaAccumulator;
    mouseDeltaAccumulator = godot::Vector2();

    for (auto i = 0; i < MAX_INPUT_INDEX; i++)
    {
        if (!input->is_action_pressed(inputKeys[i]))
            inputState[i] = false;
    }

    float mouseSensitivity = static_cast<float>(settings->getMouseSensitivity()) * 0.001;

    if (mouseDelta.length_squared() > 0.01)
    {
        attachment->rotate_object_local({0, 1, 0},
                                        mouseSensitivity * mouseDelta.x * delta * -1.0f);

        if (gravity)
        {
            head->rotate_object_local({1, 0, 0},
                                      mouseSensitivity * mouseDelta.y * delta * -1.0f);
        }
        else
        {
            attachment->rotate_object_local({1, 0, 0},
                                            mouseSensitivity * mouseDelta.y * delta * -1.0f);
        }
    }

    if (rotatePressed && continuousEnabled)
    {
        attachment->rotate_object_local({0, 1, 0}, pressedRotation.x * delta * -3.0f);

        if (gravity)
        {
            head->rotate_object_local({1, 0, 0}, pressedRotation.y * delta * -3.0f);
        }
        else
        {
            attachment->rotate_object_local({1, 0, 0}, pressedRotation.y * delta * -3.0f);
        }
    }

    godot::Vector3 movement {0, 0, 0};

    if (viewCamera->isOrthogonal())
    {
        if (inputState[INPUT_FORWARD])
            movement += {0, 1, 0};
        if (inputState[INPUT_BACKWARD])
            movement += {0, -1, 0};
        if (inputState[INPUT_LEFT])
            movement += {-1, 0, 0};
        if (inputState[INPUT_RIGHT])
            movement += {1, 0, 0};

        float newSize = viewCamera->get_size();

        if (inputState[INPUT_UP])
            newSize -= delta;
        if (inputState[INPUT_DOWN])
            newSize += delta;

        newSize -= forwardMovement * delta;
        forwardMovement = 0.0f;
        viewCamera->set_size(std::max(newSize, 0.001f));

        movement = avi::core::utilities::normalized_or_zero(movement);
    }
    else
    {
        if (inputState[INPUT_FORWARD])
            movement += {0, 0, -1};
        if (inputState[INPUT_BACKWARD])
            movement += {0, 0, 1};
        if (inputState[INPUT_LEFT])
            movement += {-1, 0, 0};
        if (inputState[INPUT_RIGHT])
            movement += {1, 0, 0};
        if (inputState[INPUT_UP])
            movement += {0, 1, 0};
        if (inputState[INPUT_DOWN])
            movement += {0, -1, 0};

        movement = avi::core::utilities::normalized_or_zero(movement);
        movement += head->get_transform().basis.rows[2] * -1 * forwardMovement * 2.0;
        forwardMovement = 0.0f;
    }

    if (inputState[INPUT_SPEED_UP])
        speedModifier += 1.0;
    if (inputState[INPUT_SPEED_DOWN])
        speedModifier -= 1.0;

    speedModifier = std::clamp(speedModifier, -100.0f, 200.0f);

    movement *= delta * static_cast<float>(settings->getMoveSpeedPlayful() * 0.01);
    movement *= std::pow(1.05f, speedModifier);

    smoothMovement = lerp(smoothMovement, movement, settings->getSmoothingFactorPlayful());
    attachment->translate(settings->isPlayfulInputSmoothed() ? smoothMovement : movement);

    if (!gravity)
    {
        float roll = 0;
        if (inputState[INPUT_ROLL_LEFT])
            roll += 1;
        if (inputState[INPUT_ROLL_RIGHT])
            roll -= 1;

        roll *= delta * static_cast<float>(settings->getRollSensitivityPlayful() * 0.02);
        smoothRoll = lerp(smoothRoll, roll, settings->getSmoothingFactorPlayful());
        attachment->rotate_object_local({0, 0, 1}, settings->isPlayfulInputSmoothed() ? smoothRoll : roll);
    }
}


void
avi::scene::PlayfulControlScheme::takeOver(avi::scene::ViewCamera* viewCamera,
                                           godot::Transform3D head)
{
    DETOOLIFY;
    this->viewCamera = viewCamera;

    attachment->set_transform(viewCamera->get_transform());
    viewCamera->set_transform(godot::Transform3D());
    this->head->set_transform(head);
    this->head->add_child(viewCamera);

    if (viewCamera->isOrthogonal())
    {
        advertiser->activateAdvertisements(adIDOrthographic);
        advertiser->deactivateAdvertisements(adID);
    }
    else
    {
        advertiser->activateAdvertisements(adID);
        advertiser->deactivateAdvertisements(adIDOrthographic);
    }

    this->smoothMovement = godot::Vector3();
}

godot::Transform3D
avi::scene::PlayfulControlScheme::release()
{
    head->remove_child(viewCamera);
    viewCamera->set_transform(attachment->get_transform());
    this->viewCamera = nullptr;

    if (advertiser != nullptr)
    {
        advertiser->deactivateAdvertisements(adID);
        advertiser->deactivateAdvertisements(adIDOrthographic);
    }

    return head->get_transform();
}

void
avi::scene::PlayfulControlScheme::reset()
{
    DETOOLIFY;
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::Vector3 transformedX = scene->get_transform().basis.xform(godot::Vector3(1, 0, 0));
    startTravel(scene->get_transform().rotated(transformedX, Math_PI * 0.5), 0.5f);
}

void
avi::scene::PlayfulControlScheme::focus(godot::AABB box)
{
    DETOOLIFY;
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::Transform3D transform;
    godot::Vector3 aabbCenter = box.position + box.size * 0.5f;

    float zoomValue = box.size.length() * 3;
    godot::Vector3 zoomVector =
        (aabbCenter - attachment->get_global_transform().origin).normalized();

    transform.origin = aabbCenter - zoomVector * zoomValue;
    transform = transform.looking_at(box.position + box.size * 0.5f, {0, 1, 0});

    this->cameraStartSize = viewCamera->get_size();
    this->cameraEndSize = box.size.length();
    this->updateSize = true;

    startTravel(transform, 0.5f);

    float clipPlaneValue = zoomValue * 0.01f;
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

void
avi::scene::PlayfulControlScheme::focus(godot::Vector3 point, avi::core::Element* target)
{
    DETOOLIFY;
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::AABB box = target->getVolume();
    godot::Transform3D transform;

    float zoomValue = box.size.length() * 3;
    godot::Vector3 zoomVector = (point - attachment->get_global_transform().origin).normalized();

    transform.origin = point - zoomVector * zoomValue;
    transform = transform.looking_at(box.position + box.size * 0.5f, {0, 1, 0});

    this->cameraStartSize = viewCamera->get_size();
    this->cameraEndSize = box.size.length() * 2;
    this->updateSize = true;

    startTravel(transform, 0.5f);

    float clipPlaneValue = zoomValue * 0.01f;
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

godot::Basis
avi::scene::PlayfulControlScheme::getRotationBasis()
{
    godot::Transform3D transform = attachment->get_transform() * head->get_transform();

    return transform.basis;
}

bool
avi::scene::PlayfulControlScheme::isActive()
{
    return viewCamera != nullptr;
}

void
avi::scene::PlayfulControlScheme::startTravel(godot::Transform3D target, float duration)
{
    DETOOLIFY;
    godot::Transform3D transformedTarget = target;
    godot::Transform3D transformHead;


    if (gravity)
    {
        float angleZ = transformedTarget.get_basis().get_euler().z;
        transformedTarget = transformedTarget.rotated(transformedTarget.get_basis().get_column(2), -angleZ);

        float angleX = transformedTarget.get_basis().get_euler().x;
        transformedTarget = transformedTarget.rotated(transformedTarget.get_basis().get_column(0), -angleX);
        transformHead.rotate({1, 0, 0}, angleX);

        transformedTarget.origin = target.origin;

        godot::Vector3 euler = transformedTarget.get_basis().get_euler();
        euler.z = 0;
        transformedTarget.basis.set_euler(euler);
    }

    travel = {attachment->get_transform() * head->get_transform(),
              std::move(transformedTarget),
              transformHead,
              duration,
              duration};
    isTraveling = true;
}

void
avi::scene::PlayfulControlScheme::doTravel(float delta)
{
    DETOOLIFY;
    travel.remainingTime -= delta;
    float progress = 1.0f - (travel.remainingTime / travel.duration);

    if (travel.remainingTime < 0)
    {
        isTraveling = false;
        updateSize = false;
        progress = 1.0f;
    }

    godot::Transform3D transform =
        travel.startTransform.interpolate_with(travel.endTransform, progress);

    if (this->updateSize || !isTraveling)
    {
        viewCamera->set_size(
            godot::Math::lerp(this->cameraStartSize, this->cameraEndSize, progress));
    }

    godot::Transform3D transformHead =
        godot::Transform3D().interpolate_with(travel.endTransformHead, progress);

    head->set_transform(transformHead);

    attachment->set_transform(transform);
}
void
avi::scene::PlayfulControlScheme::placeOnAxisNumpad(Axis axis, AxisSign sign)
{
    DETOOLIFY;
    godot::Vector3 origin = attachment->get_transform().origin;
    godot::Vector3 transformedX = scene->get_transform().basis.xform(godot::Vector3(1, 0, 0));
    godot::Transform3D transform = scene->get_transform().rotated(transformedX, Math_PI * 0.5);

    if (axis == xAxis)
    {
        godot::Basis xBasis = godot::Basis(0, 0, -1, 0, 1, 0, 1, 0, 0);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= xBasis;
                break;
            case AxisSign::negative:
                transform.basis *= xBasis.inverse();
                break;
            case AxisSign::signSwitch:
                if (switchX)
                {
                    transform.basis *= xBasis.inverse();
                }
                else
                {
                    transform.basis *= xBasis;
                }
                switchX = !switchX;
                break;
        }
    }

    if (axis == yAxis)
    {
        godot::Basis yBasis = godot::Basis(-1, 0, 0, 0, 1, 0, 0, 0, -1);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= yBasis;
                break;
            case AxisSign::negative:
                break;
            case AxisSign::signSwitch:
                if (!switchY)
                {
                    transform.basis *= yBasis;
                }
                switchY = !switchY;
                break;
        }
    }

    if (axis == zAxis)
    {
        godot::Basis zBasis = godot::Basis(1, 0, 0, 0, 0, 1, 0, -1, 0);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= zBasis;
                break;
            case AxisSign::negative:
                transform.basis *= zBasis.inverse();
                break;
            case AxisSign::signSwitch:
                if (switchZ)
                {
                    transform.basis *= zBasis.inverse();
                }
                else
                {
                    transform.basis *= zBasis;
                }
                switchZ = !switchZ;
                break;
        }
    }

    transform.set_origin(origin);
    startTravel(transform, 0.3f);
}

void
avi::scene::PlayfulControlScheme::placeOnAxis(godot::Vector3 fromVector)
{
    DETOOLIFY;
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::Transform3D transform = attachment->get_transform() * head->get_transform();

    godot::Vector3 axis = fromVector.cross(godot::Vector3(0, 0, -1)).normalized();
    float angle = Math_PI - fromVector.angle_to(godot::Vector3(0, 0, -1));

    godot::Quaternion quaternion(axis.x * sin(angle * 0.5f),
                           axis.y * sin(angle * 0.5f),
                           axis.z * sin(angle * 0.5f),
                           cos(angle * 0.5f));

    godot::Transform3D rotation(quaternion);
    transform = transform * rotation;


    transform.origin = attachment->get_transform().origin;
    startTravel(transform, 0.3f);
}

void
avi::scene::PlayfulControlScheme::invertAxis(godot::Vector3 fromVector)
{
    DETOOLIFY;
    godot::Transform3D transform = attachment->get_transform() * head->get_transform();

    float angleOffsetY =
        -0.5 * std::atan2((fromVector.cross(godot::Vector3(0, 0, -1))).dot(godot::Vector3(1, 0, 0)),
                          fromVector.dot(godot::Vector3(0, 0, -1)));
    godot::Vector3 axis = godot::Vector3(0, 1, 0).rotated(godot::Vector3(1, 0, 0), angleOffsetY);
    float signedAngle =
        std::atan2((fromVector.cross(godot::Vector3(0, 0, -1))).dot(godot::Vector3(0, 1, 0)),
                   fromVector.dot(godot::Vector3(0, 0, -1)));
    float angle = Math_PI - signedAngle;

    godot::Quaternion quaternion(axis.x * sin(angle * 0.5f),
                           axis.y * sin(angle * 0.5f),
                           axis.z * sin(angle * 0.5f),
                           cos(angle * 0.5f));
    godot::Transform3D rotation(quaternion);

    transform = transform * rotation;
    transform.origin = attachment->get_transform().origin;

    startTravel(transform, 0.3f);
}


void
avi::scene::PlayfulControlScheme::rotateFixed(RotationDirection direction, HotkeyRotation hotkey)
{
    DETOOLIFY;
    double rotation = Math_PI * 0.0833;

    if (hotkey == CameraControlScheme::disable)
    {
        continuousEnabled = true;
        rotatePressed = false;
    }

    if (hotkey == CameraControlScheme::continuous)
    {
        rotatePressed = true;

        if (direction == NEGATIVE_HORIZONTAL || direction == NEGATIVE_VERTICAL)
        {
            rotation = -rotation;
        }

        if (direction == NEGATIVE_HORIZONTAL || direction == POSITIVE_HORIZONTAL)
        {
            pressedRotation = godot::Vector2(rotation, 0);
        }
        else
        {
            pressedRotation = godot::Vector2(0, rotation);
        }
    }

    if (hotkey == CameraControlScheme::fixed)
    {
        continuousEnabled = false;

        if (direction == NEGATIVE_HORIZONTAL || direction == NEGATIVE_VERTICAL)
        {
            rotation = -rotation;
        }

        if (direction == NEGATIVE_HORIZONTAL || direction == POSITIVE_HORIZONTAL)
        {
            attachment->rotate_object_local({0, 1, 0}, -rotation);
        }
        else
        {
            if (gravity)
            {
                head->rotate_object_local({1, 0, 0}, rotation);
            }
            else
            {
                attachment->rotate_object_local({1, 0, 0}, rotation);
            }
        }
    }
}

void
avi::scene::PlayfulControlScheme::setupControlAdvertisements(
    avi::core::ControlAdvertiser& advertiser)
{
    DETOOLIFY;
    this->advertiser = &advertiser;

    using Ad = avi::core::ControlAdvertiser::Advertisement;

    {
        std::vector<avi::core::ControlAdvertiser::Advertisement> ads;
        ads.emplace_back(Ad::fromString("Playful Camera"));
        ads.emplace_back(Ad::fromString("WASD", "Move"));

        if (!gravity)
        {
            ads.emplace_back(Ad::fromString("QE", "Roll"));
        }
        ads.emplace_back(Ad::fromString("Space", "Up"));
        ads.emplace_back(Ad::fromString("X", "Down"));

        ads.emplace_back(Ad::fromString("+-", "Change Speed"));

        this->adID = advertiser.registerAdvertisement(std::move(ads));
    }

    {
        std::vector<avi::core::ControlAdvertiser::Advertisement> adsOrthographic;
        adsOrthographic.emplace_back(Ad::fromString("Playful Camera"));
        adsOrthographic.emplace_back(Ad::fromString("WASD", "Move"));
        if (!gravity)
        {
            adsOrthographic.emplace_back(Ad::fromString("QE", "Roll"));
        }
        adsOrthographic.emplace_back(Ad::fromString("Mouse Wheel", "Zoom"));
        adsOrthographic.emplace_back(Ad::fromString("+-", "Change Speed"));

        this->adIDOrthographic = advertiser.registerAdvertisement(std::move(adsOrthographic));
    }

    if (this->isActive())
    {
        advertiser.activateAdvertisements(adID);
    }
}


void
avi::scene::PlayfulControlScheme::updateControlAdvertisements()
{
    DETOOLIFY;
    if (viewCamera->isOrthogonal())
    {
        advertiser->activateAdvertisements(adIDOrthographic);
        advertiser->deactivateAdvertisements(adID);
    }
    else
    {
        advertiser->activateAdvertisements(adID);
        advertiser->deactivateAdvertisements(adIDOrthographic);
    }
}
