#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "RobotElementNode.h"

#include <VirtualRobot/CollisionDetection/CollisionModel.h>
#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/Visualization/VisualizationNode.h>

#include <godot_cpp/classes/box_mesh.hpp>
#include <godot_cpp/classes/main_loop.hpp>
#include <godot_cpp/classes/concave_polygon_shape3d.hpp>

#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

avi::scene::RobotElementNode::RobotElementNode() = default;

void
avi::scene::RobotElementNode::_bind_methods()
{
    DETOOLIFY;
    godot::ClassDB::bind_method(godot::D_METHOD("_notification"), &RobotElementNode::_notification);
}

godot::AABB
avi::scene::RobotElementNode::getVolume()
{
    bool noBaseAABB = true;
    godot::AABB AABB;
    godot::Vector3 robotScale = staticBody->get_scale() * 0.75;

    for (auto pair : nodes)
    {
        if (noBaseAABB)
        {
            AABB = pair.second->get_aabb();
            AABB.position = pair.second->get_global_transform().origin;
            AABB.size *= robotScale;
            noBaseAABB = false;
        }
        else
        {
            godot::AABB t = pair.second->get_aabb();
            t.position = pair.second->get_global_transform().origin;
            t.size *= robotScale;
            AABB = AABB.merge(t);
        }
    }
    AABB.size *= 0.5;
    AABB.position += AABB.size;

    return AABB;
}

void
avi::scene::RobotElementNode::setPosition(godot::Vector3 position)
{
    DETOOLIFY;
    auto transform = staticBody->get_transform();
    transform.set_origin(position);
    staticBody->set_transform(transform);
}

void
avi::scene::RobotElementNode::setScale(godot::Vector3 scale)
{
    DETOOLIFY;
    staticBody->set_scale(scale);
}

void
avi::scene::RobotElementNode::setRotation(godot::Quaternion rotation)
{
    DETOOLIFY;
    godot::Transform3D quaternionTransform = godot::Transform3D(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::RobotElementNode::setStartAttributes(avi::scene::GraphicsController* controller,
                                                 avi::core::ModelLoader* modelLoader,
                                                 avi::connection::SentElementType type,
                                                 std::string name,
                                                 avi::core::Layer* layer,
                                                 int flag,
                                                 std::vector<std::string> interactions,
                                                 godot::Vector3 position,
                                                 godot::Vector3 scale,
                                                 godot::Quaternion rotation,
                                                 godot::Color color,
                                                 std::string project,
                                                 std::string filename,
                                                 ModelDrawStyle drawStyle,
                                                 std::map<std::string, float> jointValues)
{
    DETOOLIFY;
    this->graphicsController = controller;
    this->modelLoader = modelLoader;

    setName(name);
    setLayer(layer);
    setType(type);

    updateAttributes(flag,
                     interactions,
                     position,
                     scale,
                     rotation,
                     color,
                     project,
                     filename,
                     drawStyle,
                     jointValues);

    controller->registerChangeListener(this, [this]() { this->visualizationChanged(); });
}

// Loading and updating robot from RobotAPI/components/ArViz/Coin/VisualizationRobot.cpp
void
avi::scene::RobotElementNode::updateAttributes(int flag,
                                               std::vector<std::string> interactions,
                                               godot::Vector3 position,
                                               godot::Vector3 scale,
                                               godot::Quaternion rotation,
                                               godot::Color color,
                                               std::string project,
                                               std::string filename,
                                               ModelDrawStyle drawStyle,
                                               const std::map<std::string, float>& jointValues)
{
    DETOOLIFY;
    setInteractionFlag(flag);
    setInteractions(interactions);
    setPosition(position);
    setRotation(rotation);
    setScale(scale * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    bool robotChanged = this->project != project || this->filename != filename;

    this->project = project;
    this->filename = filename;

    if (robotChanged or (robot == nullptr and not robotLoadingFailed))
    {
        robot = this->loadRobot(project, filename);
        if (robot != nullptr)
        {
            robot->setPropagatingJointValuesEnabled(true);
        }
        else
        {
            ARMARX_WARNING << "Error loading robot from '" << filename << "'. "
                           << "Will not try to load the robot file again until it has changed.";
            robotLoadingFailed = true;
        }
    }

    if (robot == nullptr)
    {
        return;
    }

    if (robotChanged)
    {
        visualizeRobot();
    }

    VirtualRobot::Robot& ro = *robot;
    for (auto& [nodeName, newJointValue] : jointValues)
    {
        VirtualRobot::RobotNodePtr robotNode = ro.getRobotNode(nodeName);
        if (robotNode == nullptr)
        {
            continue;
        }

        const float oldJointValue = robotNode->getJointValue();
        const float diff = std::abs(newJointValue - oldJointValue);
        const bool jointValuesChanged = diff > 0.001f;
        if (jointValuesChanged)
        {
            // Only set the joint values if they changed.
            // This call causes internal updates to the visualization even if joint angles are the same!
            ro.setJointValues(jointValues);
            break;
        }
    }

    if (this->armarXDrawStyle != drawStyle || this->color != color)
    {
        this->armarXDrawStyle = drawStyle;
        this->color = color;

        isOverrideDirty = true;
    }

    visualizationChanged();
}

void
avi::scene::RobotElementNode::updateVisualization()
{
    DETOOLIFY;
    VirtualRobot::Robot& ro = *robot;

    for (const auto& nodeName : ro.getRobotNodeNames())
    {
        if (nodes.find(nodeName) != nodes.end())
        {
            auto node = ro.getRobotNode(nodeName);
            const Eigen::Matrix4f nodePosition = node->getVisualization()
                                                     ? node->getVisualization()->getGlobalPose()
                                                     : node->getGlobalPose();

            godot::Vector3 x = {nodePosition(0, 0), nodePosition(1, 0), nodePosition(2, 0)};
            godot::Vector3 y = {nodePosition(0, 1), nodePosition(1, 1), nodePosition(2, 1)};
            godot::Vector3 z = {nodePosition(0, 2), nodePosition(1, 2), nodePosition(2, 2)};

            godot::Vector3 origin = {nodePosition(0, 3) ,
                                     nodePosition(1, 3) ,
                                     nodePosition(2, 3) };
            origin /= avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;

            godot::Transform3D transform = godot::Transform3D({x, y, z}, origin);

            nodes.at(nodeName)->set_transform(transform);

            if (collisionShapes.find(nodeName) != collisionShapes.end())
            {
                collisionShapes.at(nodeName)->set_transform(transform);
            }
        }
    }
}

avi::scene::RobotElementNode*
avi::scene::RobotElementNode::newRobotElementNode()
{
    RobotElementNode* node = memnew(avi::scene::RobotElementNode);
    node->staticBody = memnew(godot::StaticBody3D);
    node->add_child(node->staticBody);
    return node;
}

VirtualRobot::RobotPtr
avi::scene::RobotElementNode::loadRobot(std::string project, std::string filename)
{
    VirtualRobot::RobotPtr result;

    auto cacheEntry = getCache().get(project, filename);
    if (cacheEntry.has_value())
    {
        return cacheEntry.value();
    }

    if (filename.empty())
    {
        ARMARX_WARNING << "Error loading robot: filename is emtpy";
        return result;
    }

    if (project.empty() && std::filesystem::path(filename).is_absolute())
    {
        ARMARX_INFO << deactivateSpam() << "You specified the absolute path to the robot file:"
                    << "\n\t'" << filename << "'"
                    << "\nConsider specifying the containing ArmarX package and relative data path "
                       "instead to "
                    << "improve portability to other systems.";
    }
    std::string fullFilename;
    armarx::ArmarXDataPath::FindPackageAndAddDataPath(project);

    if (!armarx::ArmarXDataPath::SearchReadableFile(filename, fullFilename))
    {
        ARMARX_WARNING << "Robot could not be loaded";
        return result;
    }

    VirtualRobot::RobotIO::RobotDescription loadMode = VirtualRobot::RobotIO::eFull;

    try
    {
        ARMARX_INFO << "Loading robot from " << fullFilename;
        result = VirtualRobot::RobotIO::loadRobot(fullFilename, loadMode);
        if (result)
        {
            result->setThreadsafe(false);
            result->setPropagatingJointValuesEnabled(false);
        }
        else
        {
            ARMARX_WARNING << "Could not load robot from file: " << fullFilename;
        }
    }
    catch (VirtualRobot::VirtualRobotException& e)
    {
        ARMARX_WARNING << "Could not load robot from file: " << fullFilename
                       << "\nReason: " << e.what();
    }

    this->project = project;
    this->filename = filename;

    getCache().set(project, filename, result);

    return result;
}

void
avi::scene::RobotElementNode::visualizeRobot()
{
    DETOOLIFY;
    isOverrideDirty = true;

    for (auto node : nodes)
    {
        node.second->queue_free();
    }

    nodes.clear();

    for (const auto& collisionShape : collisionShapes)
    {
        collisionShape.second->queue_free();
    }

    collisionShapes.clear();

    if (!robot)
    {
        return;
    }

    for (const std::string& nodeName : robot->getRobotNodeNames())
    {
        auto node = robot->getRobotNode(nodeName);

        auto loadObject = [&](VirtualRobot::VisualizationNodePtr visualization)
            -> std::optional<godot::Ref<godot::Mesh>>
        {
            std::string nodeFilename = visualization->getFilename();

            if (not nodeFilename.empty())
            {
                godot::Ref<godot::Mesh> object = modelLoader->loadModelFromFile(nodeFilename);
                return object;
            }

            if (visualization->getTriMeshModel())
            {
                godot::Ref<godot::Mesh> object = avi::core::ModelLoader::createObjectFromTriMesh(
                    visualization->getTriMeshModel());

                return object;
            }

            return {};
        };

        auto addMeshFromObject = [&](std::optional<godot::Ref<godot::Mesh>>& object)
        {
            if (not object.has_value())
                return;

            auto mesh = object.value();
            godot::MeshInstance3D* m = memnew(godot::MeshInstance3D);
            m->set_name(nodeName.c_str());
            m->set_mesh(mesh);

            staticBody->add_child(m);
            nodes.emplace(nodeName, m);
        };

        auto addShapeFromObject = [&](std::optional<godot::Ref<godot::Mesh>>& object)
        {
            if (not object.has_value())
            {
                return;
            }

            auto mesh = object.value();

            addShape(mesh, nodeName);
        };

        VirtualRobot::VisualizationNodePtr visualization = node->getVisualization();
        VirtualRobot::VisualizationNodePtr collision =
            node->getCollisionModel() ? node->getCollisionModel()->getVisualization() : nullptr;

        bool useCollisionAsVisualization = robotDrawStyle != avi::scene::DrawStyle::VISUALIZATION;

        if (useCollisionAsVisualization)
        {
            if (collision)
            {
                auto object = loadObject(collision);
                addMeshFromObject(object);
                addShapeFromObject(object);
            }
        }
        else
        {
            if (visualization)
            {
                auto object = loadObject(visualization);
                addMeshFromObject(object);
            }

            if (collision)
            {
                auto object = loadObject(collision);
                addShapeFromObject(object);
            }
        }
    }
}

void
avi::scene::RobotElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    DETOOLIFY;
    auto robot = IceInternal::Handle<armarx::viz::data::ElementRobot>::dynamicCast(elementInfo);

    updateAttributes(robot->interaction.enableFlags,
                     robot->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(robot->pose),
                     avi::scene::Converter::getScale(robot->scale),
                     avi::scene::Converter::getRotation(robot->pose),
                     avi::scene::Converter::getColor(robot->color),
                     robot->project,
                     robot->filename,
                     ModelDrawStyle(robot->drawStyle),
                     robot->jointValues);
}

godot::Transform3D
avi::scene::RobotElementNode::getTransform()
{
    return staticBody->get_transform();
}

void
avi::scene::RobotElementNode::addShape(godot::Ref<godot::Mesh> mesh, const std::string& nodeName)
{
    DETOOLIFY;
    godot::CollisionShape3D* collisionShape = memnew(godot::CollisionShape3D);

    std::optional<godot::Ref<godot::Shape3D>> shape = modelLoader->getShapeCache().get(mesh.ptr());

    if (shape.has_value())
    {
        collisionShape->set_shape(shape.value());
    }
    else
    {
        godot::Ref<godot::Shape3D> newShape = mesh->create_trimesh_shape();
        modelLoader->getShapeCache().emplace(mesh.ptr(), {newShape});

        collisionShape->set_shape(newShape);
    }

    staticBody->add_child(collisionShape);
    collisionShapes.emplace(nodeName, collisionShape);
}

void
avi::scene::RobotElementNode::visualizationChanged()
{
    DETOOLIFY;
    if (!robot)
    {
        robot = this->loadRobot(project, filename);
    }

    if (!robot)
    {
        ARMARX_ERROR << "Error loading robot from '" << filename << "'";
        return;
    }

    bool drawStyleChanged = false;
    avi::scene::DrawStyle sceneDrawStyle = graphicsController->getCurrentDrawStyle();
    if (sceneDrawStyle == avi::scene::DrawStyle::ORIGINAL)
    {
        bool isUsingVisualization = robotDrawStyle == avi::scene::DrawStyle::VISUALIZATION;
        bool isRequestingVisualization = armarXDrawStyle.isVisualization();

        if (isUsingVisualization != isRequestingVisualization)
        {
            drawStyleChanged = true;
            robotDrawStyle = armarXDrawStyle.getSceneDrawStyle();
        }
    }
    else if (sceneDrawStyle != robotDrawStyle)
    {
        drawStyleChanged = true;
        robotDrawStyle = sceneDrawStyle;
    }

    if (drawStyleChanged)
    {
        visualizeRobot();
    }

    updateVisualization();
    setOverrideColor(this->color, this->armarXDrawStyle);
}

void
avi::scene::RobotElementNode::_notification(int what)
{
    DETOOLIFY;
    if (what == godot::MainLoop::NOTIFICATION_PREDELETE && graphicsController != nullptr)
    {
        graphicsController->unregisterChangeListener(this);
    }
}

std::vector<godot::Node3D*>
avi::scene::RobotElementNode::getSpatial()
{
    std::vector<godot::Node3D*> vec;
    for (auto node : nodes)
    {
        vec.push_back(node.second);
    }
    return vec;
}

void
avi::scene::RobotElementNode::setOverrideColor(godot::Color color, ModelDrawStyle drawStyle)
{
    DETOOLIFY;
    if (not isOverrideDirty)
    {
        return;
    }

    if (drawStyle.isColorOverriden())
    {
        if (not nodes.empty())
        {
            auto& [key, value] = *nodes.begin();
            godot::Ref<godot::StandardMaterial3D> material;

            if (value->get_material_override().is_null())
            {
                material = godot::Ref(memnew(godot::StandardMaterial3D));
            }
            else
            {
                material = value->get_material_override();
            }

            material->set_albedo(color);
            material->set_flag(godot::StandardMaterial3D::FLAG_ALBEDO_FROM_VERTEX_COLOR, true);

            if (color.a != 1)
            {
                material->set_blend_mode(godot::StandardMaterial3D::BLEND_MODE_MIX);
                material->set_transparency(godot::StandardMaterial3D::TRANSPARENCY_ALPHA);
            }

            for (auto& node : nodes)
            {
                node.second->set_material_override(material);
            }
        }
    }
    else
    {
        for (auto& node : nodes)
        {
            node.second->set_material_override(nullptr);
        }
    }
}

avi::scene::RobotElementNode::RobotCache&
avi::scene::RobotElementNode::getCache()
{
    static RobotCache cache;
    return cache;
}

std::optional<VirtualRobot::RobotPtr>
avi::scene::RobotElementNode::RobotCache::get(const std::string& project,
                                              const std::string& filename)
{
    Key searched = std::make_pair(project, filename);
    size_t idx = 0;

    for (idx = 0; idx < entries.size(); idx++)
    {
        Entry& entry = entries[idx];

        if (entry.key == searched)
        {
            break;
        }
    }

    if (idx < entries.size())
    {
        Entry entry = std::move(entries[idx]);

        entries.emplace_back(std::move(entry));
        entries.erase(entries.begin() + idx);

        return entries.back().value;
    }

    return {};
}

void
avi::scene::RobotElementNode::RobotCache::set(const std::string& project,
                                              const std::string& filename,
                                              VirtualRobot::RobotPtr robot)
{
    DETOOLIFY;
    if (entries.size() >= 8)
    {
        entries.erase(entries.begin());
    }

    entries.push_back({{project, filename}, std::move(robot)});
}
