#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "TechnicalControlScheme.h"

#include <cmath>

#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/input.hpp>
#include <godot_cpp/classes/input_event_mouse_button.hpp>
#include <godot_cpp/classes/input_event_mouse_motion.hpp>

void
avi::scene::TechnicalControlScheme::_bind_methods()
{
    DETOOLIFY;


}

avi::scene::TechnicalControlScheme::TechnicalControlScheme()
{
    DETOOLIFY;
    mouseDelta = {0, 0};
    wheelDelta = 0;
    smoothWheelDelta = 0;

    zoomValue = 0.0f;
    zoomDistance = 0.0f;

    viewCamera = nullptr;
}

void
avi::scene::TechnicalControlScheme::setup(avi::core::SettingsProvider& settings,
                                          avi::scene::VisualizedScene& scene,
                                          bool gravity)
{
    this->settings = &settings;
    this->scene = &scene;
    this->gravity = gravity;
}

void
avi::scene::TechnicalControlScheme::_ready()
{
    DETOOLIFY;
    anchor = get_node<godot::Node3D>("Anchor");
    head = get_node<godot::Node3D>("Anchor/Head");
    zoomer = get_node<godot::Node3D>("Anchor/Head/Zoom");

    attachment = get_node<godot::Node3D>("Anchor/Head/Zoom/Attachment");
}

void
avi::scene::TechnicalControlScheme::_unhandled_input(const godot::Ref<godot::InputEvent>& event)
{
    DETOOLIFY;
    if (!isActive()) return;

    auto* mouseMotionEvent = cast_to<godot::InputEventMouseMotion>(event.ptr());
    if (mouseMotionEvent)
    {
        mouseDelta = mouseMotionEvent->get_relative();
    }

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event.ptr());
    if (mouseButtonEvent)
    {
        float change = 0.0f;
        auto idx = mouseButtonEvent->get_button_index();

        if (idx == godot::MOUSE_BUTTON_WHEEL_UP)
        {
            change += 1.0f;
        }

        if (idx == godot::MOUSE_BUTTON_WHEEL_DOWN)
        {
            change -= 1.0f;
        }

        wheelDelta += change;
    }
}

void
avi::scene::TechnicalControlScheme::_process(double delta)
{
    DETOOLIFY;
    if (!isActive()) return;

    godot::Vector2 oldMouseDelta = mouseDelta;
    mouseDelta = {0, 0};

    float oldWheelDelta = wheelDelta;
    wheelDelta = 0;

    smoothWheelDelta = godot::Math::lerp(smoothWheelDelta, oldWheelDelta, 0.4f);

    if (isTraveling)
    {
        doTravel(delta);
        return;
    }

    godot::Input* input = godot::Input::get_singleton();

    // Whether the camera should be moved in any way
    bool moveCamera = input->is_action_pressed("avi_camera_mouse") || viewCamera->isMouseLocked();
    // Whether the camera should use alternative movement (e.g. move anchor instead of around anchor)
    bool alternative = input->is_action_pressed("avi_camera_alt");

    if (moveCamera)
    {
        if (!alternative)
        {
            orbit(oldMouseDelta, delta);
        }
        else
        {
            move(oldMouseDelta, delta);
        }
    }

    if (!alternative)
    {
        if (viewCamera->isOrthogonal())
        {
            zoomOrthographic(smoothWheelDelta, delta);
        }
        else
        {
            zoom(smoothWheelDelta, delta);
        }
    }
    else
    {
        roll(smoothWheelDelta, delta);
    }

    if (rotatePressed && continuousEnabled)
    {
        orbit(pressedRotation, delta * 15);
    }

    anchor->set_transform(anchor->get_transform().orthonormalized());
    head->set_transform(head->get_transform().orthonormalized());
    zoomer->set_transform(zoomer->get_transform().orthonormalized());
}

void
avi::scene::TechnicalControlScheme::takeOver(avi::scene::ViewCamera* viewCamera,
                                             godot::Transform3D head)
{
    DETOOLIFY;
    this->viewCamera = viewCamera;
    godot::Transform3D change =
        viewCamera->get_transform() * head * zoomer->get_transform().inverse() * head.inverse();
    viewCamera->set_transform(godot::Transform3D());
    anchor->set_transform(change);
    this->head->set_transform(head);

    attachment->add_child(viewCamera);

    if (advertiser != nullptr)
    {
        advertiser->activateAdvertisements(adID);
    }
}

godot::Transform3D
avi::scene::TechnicalControlScheme::release()
{
    attachment->remove_child(viewCamera);
    viewCamera->set_transform(anchor->get_transform() * this->head->get_transform() *
                              zoomer->get_transform() * this->head->get_transform().inverse());
    this->viewCamera = nullptr;

    advertiser->deactivateAdvertisements(adID);

    return head->get_transform();
}

void
avi::scene::TechnicalControlScheme::reset()
{
    DETOOLIFY;
    anchor->set_transform(anchor->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());
    godot::Vector3 transformedX = scene->get_transform().basis.xform(godot::Vector3(1, 0, 0));
    startTravel(scene->get_transform().rotated(transformedX, Math_PI * 0.5), 0.0f, 0.5f);
}

void
avi::scene::TechnicalControlScheme::focus(godot::AABB box)
{
    DETOOLIFY;
    anchor->set_transform(anchor->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());
    godot::Transform3D transform = anchor->get_transform();
    transform.origin = box.position + box.size * 0.5f;

    float distance = box.size.length() * 3;
    float value = -std::log(distance);

    this->cameraStartSize = viewCamera->get_size();
    this->cameraEndSize = box.size.length();
    this->updateSize = true;

    startTravel(transform, value, 0.5f);
}

void
avi::scene::TechnicalControlScheme::focus(godot::Vector3 point, avi::core::Element* target)
{
    DETOOLIFY;
    anchor->set_transform(anchor->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());
    godot::Transform3D transform = anchor->get_transform();
    transform.origin = point;

    float distance = target->getVolume().size.length() * 3;
    float value = -std::log(distance);

    this->cameraStartSize = viewCamera->get_size();
    this->cameraEndSize = target->getVolume().size.length() * 2;
    this->updateSize = true;

    startTravel(transform, value, 0.5f);
}

godot::Basis
avi::scene::TechnicalControlScheme::getRotationBasis()
{
    return (anchor->get_transform() * head->get_transform() * zoomer->get_transform().inverse())
        .basis.inverse();
}

bool
avi::scene::TechnicalControlScheme::isActive()
{
    return viewCamera != nullptr;
}

void
avi::scene::TechnicalControlScheme::orbit(godot::Vector2 rotation, float deltaTime)
{
    DETOOLIFY;
    const float sensitivity = static_cast<float>(settings->getOrbitSensitivity() * 0.001);
    rotation *= sensitivity * deltaTime;

    if (gravity)
    {
        anchor->rotate_object_local({0, 1, 0}, -rotation.x);
        head->rotate_object_local({1, 0, 0}, -rotation.y);
    }
    else
    {
        anchor->rotate_object_local({0, 1, 0}, -rotation.x);
        anchor->rotate_object_local({1, 0, 0}, -rotation.y);
    }
}

void
avi::scene::TechnicalControlScheme::rotateFixed(RotationDirection direction, HotkeyRotation hotkey)
{
    DETOOLIFY;
    double rotation = Math_PI * 0.0833;

    if (hotkey == CameraControlScheme::disable)
    {
        continuousEnabled = true;
        rotatePressed = false;
    }

    if (hotkey == CameraControlScheme::continuous)
    {
        rotatePressed = true;

        if (direction == NEGATIVE_HORIZONTAL || direction == NEGATIVE_VERTICAL)
        {
            rotation = -rotation;
        }

        if (direction == NEGATIVE_HORIZONTAL || direction == POSITIVE_HORIZONTAL)
        {
            pressedRotation = godot::Vector2(-rotation, 0);
        }
        else
        {
            pressedRotation = godot::Vector2(0, -rotation);
        }
    }

    if (hotkey == CameraControlScheme::fixed)
    {
        continuousEnabled = false;

        if (direction == NEGATIVE_HORIZONTAL || direction == NEGATIVE_VERTICAL)
        {
            rotation = -rotation;
        }

        if (direction == NEGATIVE_HORIZONTAL || direction == POSITIVE_HORIZONTAL)
        {
            anchor->rotate_object_local({0, 1, 0}, rotation);
        }
        else
        {
            anchor->rotate_object_local({1, 0, 0}, -rotation);
        }
    }
}


void
avi::scene::TechnicalControlScheme::move(godot::Vector2 movement, float deltaTime)
{
    DETOOLIFY;
    const float sensitivity = static_cast<float>(settings->getMoveSpeedTechnical() * 0.01);
    movement *= sensitivity * deltaTime;
    movement *= zoomDistance;

    godot::Transform3D attachmentAtAnchor = anchor->get_transform() * head->get_transform() *
                                          zoomer->get_transform() * attachment->get_transform();
    godot::Basis basis = attachmentAtAnchor.basis;

    godot::Transform3D newTransform = anchor->get_transform();

    newTransform.origin += basis.get_column(0) * -movement.x;
    newTransform.origin += basis.get_column(1) * +movement.y;

    anchor->set_transform(newTransform);
}

void
avi::scene::TechnicalControlScheme::zoom(float delta, float deltaTime)
{
    DETOOLIFY;
    const float sensitivity = static_cast<float>(settings->getZoomSensitivity() * 0.01);
    delta *= sensitivity * deltaTime;

    zoomValue += delta;
    zoomValue = getClampedZoom(zoomValue);
    zoomDistance = std::exp(-1 * zoomValue);

    godot::Transform3D transform = zoomer->get_transform();
    transform.origin = godot::Vector3(0, 0, zoomDistance);
    zoomer->set_transform(transform);

    float clipPlaneValue;

    if (zoomValue > 4)
    {
        clipPlaneValue = std::pow(4, -zoomValue);
    }
    else
    {
        clipPlaneValue = std::pow(2, -(zoomValue + 4));
    }
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

void
avi::scene::TechnicalControlScheme::zoomOrthographic(float delta, float deltaTime)
{
    DETOOLIFY;
    const float sensitivity = static_cast<float>(settings->getZoomSensitivity() * 0.02);
    delta *= sensitivity * deltaTime;

    viewCamera->set_size((viewCamera->get_size() - delta < 0.001) ? 0.001
                                                                  : viewCamera->get_size() - delta);
}

void
avi::scene::TechnicalControlScheme::roll(float delta, float deltaTime)
{
    DETOOLIFY;
    if (gravity)
    {
        return;
    }

    const float sensitivity = static_cast<float>(settings->getRollSensitivity() * 0.02);
    delta *= sensitivity * deltaTime;

    anchor->rotate_object_local({0, 0, 1}, delta);
}

void
avi::scene::TechnicalControlScheme::startTravel(godot::Transform3D anchor, float zoom, float duration)
{
    DETOOLIFY;
    godot::Transform3D transformedAnchor = anchor;
    godot::Transform3D transformHead;

    if (gravity)
    {
        float angleZ = transformedAnchor.get_basis().get_euler().z;
        transformedAnchor = transformedAnchor.rotated(transformedAnchor.get_basis().get_column(2), -angleZ);

        float angleX = transformedAnchor.get_basis().get_euler().x;
        transformedAnchor = transformedAnchor.rotated(transformedAnchor.get_basis().get_column(0), -angleX);
        transformHead.rotate({1, 0, 0}, angleX);

        transformedAnchor.origin = anchor.origin;

        godot::Vector3 euler = transformedAnchor.get_basis().get_euler();
        euler.z = 0;
        transformedAnchor.basis.set_euler(euler);
    }

    travel = {.startTransform = this->anchor->get_transform(),
              .endTransform = transformedAnchor,
              .endTransformHead = transformHead,
              .startZoom = this->zoomValue,
              .endZoom = zoom,
              .duration = duration,
              .remainingTime = duration};

    isTraveling = true;
}

void
avi::scene::TechnicalControlScheme::doTravel(float deltaTime)
{
    DETOOLIFY;
    travel.remainingTime -= deltaTime;
    float progress = 1.0f - (travel.remainingTime / travel.duration);

    if (travel.remainingTime < 0)
    {
        isTraveling = false;
        updateSize = false;
        progress = 1.0f;
    }

    godot::Transform3D transform =
        travel.startTransform.interpolate_with(travel.endTransform, progress);


    if (this->updateSize || !isTraveling)
    {
        viewCamera->set_size(
            godot::Math::lerp(this->cameraStartSize, this->cameraEndSize, progress));
    }
    zoomValue = godot::Math::lerp(travel.startZoom, getClampedZoom(travel.endZoom), progress);
    zoom(0, deltaTime);

    godot::Transform3D transformHead =
        godot::Transform3D().interpolate_with(travel.endTransformHead, progress);

    head->set_transform(transformHead);
    anchor->set_transform(transform);
}

float
avi::scene::TechnicalControlScheme::getClampedZoom(float zoom)
{
    return std::clamp(zoom, -250.0f, 10.0f);
}

void
avi::scene::TechnicalControlScheme::placeOnAxisNumpad(Axis axis, AxisSign sign)
{
    DETOOLIFY;
    anchor->set_transform(anchor->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::Vector3 transformedX = scene->get_transform().basis.xform(godot::Vector3(1, 0, 0));
    godot::Transform3D transform = scene->get_transform().rotated(transformedX, Math_PI * 0.5);

    if (axis == xAxis)
    {
        godot::Basis xBasis = godot::Basis(0, 0, -1, 0, 1, 0, 1, 0, 0);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= xBasis;
                break;
            case AxisSign::negative:
                transform.basis *= xBasis.inverse();
                break;
            case AxisSign::signSwitch:
                if (switchX)
                {
                    transform.basis *= xBasis.inverse();
                }
                else
                {
                    transform.basis *= xBasis;
                }
                switchX = !switchX;
                break;
        }
    }

    if (axis == yAxis)
    {
        godot::Basis yBasis = godot::Basis(-1, 0, 0, 0, 1, 0, 0, 0, -1);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= yBasis;
                break;
            case AxisSign::negative:
                break;
            case AxisSign::signSwitch:
                if (!switchY)
                {
                    transform.basis *= yBasis;
                }
                switchY = !switchY;
                break;
        }
    }

    if (axis == zAxis)
    {
        godot::Basis zBasis = godot::Basis(1, 0, 0, 0, 0, 1, 0, -1, 0);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= zBasis;
                break;
            case AxisSign::negative:
                transform.basis *= zBasis.inverse();
                break;
            case AxisSign::signSwitch:
                if (switchZ)
                {
                    transform.basis *= zBasis.inverse();
                }
                else
                {
                    transform.basis *= zBasis;
                }
                switchZ = !switchZ;
                break;
        }
    }

    transform.origin = anchor->get_transform().origin;
    startTravel(transform, this->zoomValue, 0.3f);
}

void
avi::scene::TechnicalControlScheme::placeOnAxis(godot::Vector3 fromVector)
{
    DETOOLIFY;
    anchor->set_transform(anchor->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::Transform3D transform = anchor->get_transform();

    godot::Vector3 axis = fromVector.cross(godot::Vector3(0, 0, -1)).normalized();
    float angle = Math_PI - fromVector.angle_to(godot::Vector3(0, 0, -1));

    godot::Quaternion quaternion(axis.x * sin(angle * 0.5f),
                           axis.y * sin(angle * 0.5f),
                           axis.z * sin(angle * 0.5f),
                           cos(angle * 0.5f));

    godot::Transform3D rotation(quaternion);
    transform = transform * rotation;

    transform.origin = anchor->get_transform().origin;
    startTravel(transform, this->zoomValue, 0.3f);
}


void
avi::scene::TechnicalControlScheme::invertAxis(godot::Vector3 fromVector)
{
    DETOOLIFY;
    anchor->set_transform(anchor->get_transform() * head->get_transform());
    head->set_transform(godot::Transform3D());

    godot::Transform3D transform = anchor->get_transform();

    float angleOffsetY =
        -0.5 * std::atan2((fromVector.cross(godot::Vector3(0, 0, -1))).dot(godot::Vector3(1, 0, 0)),
                          fromVector.dot(godot::Vector3(0, 0, -1)));
    godot::Vector3 axis = godot::Vector3(0, 1, 0).rotated(godot::Vector3(1, 0, 0), angleOffsetY);
    float signedAngle =
        std::atan2((fromVector.cross(godot::Vector3(0, 0, -1))).dot(godot::Vector3(0, 1, 0)),
                   fromVector.dot(godot::Vector3(0, 0, -1)));
    float angle = Math_PI - signedAngle;

    godot::Quaternion quaternion(axis.x * sin(angle * 0.5f),
                           axis.y * sin(angle * 0.5f),
                           axis.z * sin(angle * 0.5f),
                           cos(angle * 0.5f));
    godot::Transform3D rotation(quaternion);

    transform = transform * rotation;
    transform.origin = anchor->get_transform().origin;

    startTravel(transform, this->zoomValue, 0.3f);
}


void
avi::scene::TechnicalControlScheme::setupControlAdvertisements(
    avi::core::ControlAdvertiser& advertiser)
{
    DETOOLIFY;
    this->advertiser = &advertiser;

    using Ad = avi::core::ControlAdvertiser::Advertisement;

    std::vector<avi::core::ControlAdvertiser::Advertisement> ads;
    ads.emplace_back(Ad::fromString("Technical Camera"));
    ads.emplace_back(Ad::fromString("Mouse Wheel Down", "Orbit"));
    ads.emplace_back(Ad::fromString("Mouse Wheel Scroll", "Zoom"));
    ads.emplace_back(Ad::fromString("Mouse Wheel Down + Shift", "Move"));
    if (!gravity)
    {
        ads.emplace_back(Ad::fromString("Mouse Wheel Scroll + Shift", "Roll"));
    }

    this->adID = advertiser.registerAdvertisement(std::move(ads));

    if (this->isActive())
    {
        advertiser.activateAdvertisements(adID);
    }
}
