#include <arviz_godot/lib/avi/core/Detoolifier.h>
#include "VisualizedScene.h"

void
avi::scene::VisualizedScene::_bind_methods()
{
    DETOOLIFY;
    godot::ClassDB::bind_method(godot::D_METHOD("_notification"), &VisualizedScene::_notification);

    ADD_SIGNAL(godot::MethodInfo("layers_changed"));
}

void
avi::scene::VisualizedScene::_ready()
{
    DETOOLIFY;
    coordinateSystem = get_node<godot::Node3D>("CoordinateSystem");
}

void
avi::scene::VisualizedScene::_process(double delta)
{
    DETOOLIFY;

    auto changedLayers = converter->apply();
    auto newLayers = changedLayers.layers;

    if(changedLayers.clearScene) {
        clearScene();
        return;
    }

    if (newLayers != layers)
    {
        layers = std::move(newLayers);

        std::map<std::string, Component> oldComponents = std::move(components);
        components.clear();

        // Create the new component structure.

        for (auto& [name, layer] : layers)
        {
            components[name.component].layers[name.name] = layer;
        }

        // Transfer visibility of old components to new.

        for (auto& [name, newComponent] : components)
        {
            auto cit = oldComponents.find(name);

            if (cit != oldComponents.end())
            {
                auto& [_, oldComponent] = *cit;
                newComponent.isVisible = oldComponent.isVisible;
            }
            else
            {
                newComponent.isVisible = true;
            }
        }

        // Set visibility of added layers.

        for (auto& [name, layer] : layers)
        {
            auto cit = oldComponents.find(name.component);

            if (cit != oldComponents.end())
            {
                auto& [_, oldComponent] = *cit;

                if (oldComponent.layers.find(name.name) == oldComponent.layers.end())
                {
                    layer->setVisibility(oldComponent.isVisible);
                }
            }
        }

        emit_signal("layers_changed");
    }
}

void
avi::scene::VisualizedScene::_notification(int what)
{
    DETOOLIFY;
    if (what == godot::Object::NOTIFICATION_PREDELETE)
    {
        converter->clearCaches();
    }
}

void
avi::scene::VisualizedScene::setup(avi::connection::RevisionBuffer& revisionBuffer,
                                   avi::scene::GraphicsController* graphicsController)
{
    DETOOLIFY;
    this->graphicsController = graphicsController;
    converter = std::make_unique<avi::scene::Converter>(this, revisionBuffer, graphicsController);

    this->graphicsController->registerChangeListener(this,
                                                     [this]() { this->visualizationChanged(); });
    visualizationChanged();

    setOrientation({-90, 0, 0});
}

std::vector<std::string>
avi::scene::VisualizedScene::getComponents()
{
    std::vector<std::string> componentList;

    for (const auto& [name, _] : components)
    {
        componentList.push_back(name);
    }

    return componentList;
}

std::vector<std::string>
avi::scene::VisualizedScene::getLayers(const std::string& component)
{
    std::vector<std::string> nameList;

    for (const auto& [name, layer] : components[component].layers)
    {
        nameList.push_back(name);
    }

    return nameList;
}

void
avi::scene::VisualizedScene::setLayerVisibility(const std::string& component,
                                                const std::string& layer,
                                                bool isVisible)
{
    auto cit = components.find(component);

    if (cit != components.end())
    {
        auto lit = cit->second.layers.find(layer);

        if (lit != cit->second.layers.end())
        {
            lit->second->setVisibility(isVisible);
        }
    }
}

bool
avi::scene::VisualizedScene::getLayerVisibility(const std::string& component,
                                                const std::string& layer)
{
    auto cit = components.find(component);

    if (cit != components.end())
    {
        auto lit = cit->second.layers.find(layer);

        if (lit != cit->second.layers.end())
        {
            return lit->second->isVisible();
        }
    }

    return false;
}

void
avi::scene::VisualizedScene::setComponentVisibility(const std::string& component, bool isVisible)
{
    DETOOLIFY;
    auto cit = components.find(component);

    if (cit != components.end())
    {
        cit->second.isVisible = isVisible;
    }
}

bool
avi::scene::VisualizedScene::getComponentVisibility(const std::string& component)
{
    auto cit = components.find(component);

    if (cit != components.end())
    {
        return cit->second.isVisible;
    }

    return false;
}

godot::AABB
avi::scene::VisualizedScene::getSceneVolume()
{
    bool noBaseAABB = true;
    godot::AABB aabb;

    for (auto& [name, layer] : layers)
    {
        if (layer->isVisible())
        {
            if (noBaseAABB)
            {
                aabb = layer->getAABB();
                noBaseAABB = false;
            }
            else
            {
                aabb = aabb.merge(layer->getAABB());
            }
        }
    }

    return aabb;
}

void
avi::scene::VisualizedScene::clearScene()
{
    DETOOLIFY;
    for (auto& layer : layers)
    {
        layer.second->saveRemove();
    }
    layers.clear();
    converter->clear();
}

godot::Vector3
avi::scene::VisualizedScene::getOrientation()
{
    godot::Vector3 orientation = this->get_rotation_degrees();
    return {clipDegrees(orientation.x), clipDegrees(orientation.y), clipDegrees(orientation.z)};
}

void
avi::scene::VisualizedScene::setOrientation(const godot::Vector3& orientation)
{
    DETOOLIFY;
    this->set_rotation_degrees(orientation);
}

real_t
avi::scene::VisualizedScene::clipDegrees(real_t degrees)
{
    auto degreesInt = static_cast<int64_t>(degrees);
    degreesInt = ((degreesInt % 360) + 360) % 360;

    if (degreesInt > 180)
    {
        degreesInt -= 360;
    }

    return static_cast<real_t>(degreesInt);
}

void
avi::scene::VisualizedScene::visualizationChanged()
{
    DETOOLIFY;
    coordinateSystem->set_visible(graphicsController->isCoordinateSystemEnabled());
}
